#!/usr/bin/env python3
#
# Tests the HTML support functions.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit.formats.html


html = """
<!DOCTYPE html>
<html xmlns="http://www.w3.org/1999/xhtml"
      xmlns:xhtml="http://www.w3.org/1999/xhtml"
      lang="en">
<head>
    <meta charset="utf-8" />
    <meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1" />
    <title>Myokit &amp; an entity test</title>
    <base href="http://myokit.org/" />
    <link rel="stylesheet" type="text/css" media="all"
        href="http://myokit.org/static/css/myokit-16.css" />
    <script src="http://myokit.org/static/js/rainbow-custom.min.js"
        type="text/javascript"></script>
</head>
<body>
    <!--[if lt IE 7]> <div>Don't use IE.</div><![endif]-->
    <div class="header">
        <h1>Myokit</h1>
        <p>Text example 1</p>
        <!-- Image link -->
        <a href="http://myokit.org/">
            <img src="static/img/logo.png" alt="Logo" />
        </a>
        <!-- Text link -->
        <a href="link">This should be visible</a>
        <p>More text.</p>
        <p>Issue
    </div>
    <div>
        <h2>Software</h2>
        <strong>This is strong</strong> &amp; this is an ampersand.
        <hr />
        <ul>
            <li>One</li>
            <li><a href="http://myokit.org/download">Two</a></li>
        </ul>
        <h3>Resources</h3>
        This has a<xhtml:br />
        line break.
        <ol>
            <xhtml:li>Numbered A.</xhtml:li>
            <li><em>Numbered B is italic.</em></li>
            <li>
                <ul>
                    <li>C is double indented</li>
                    <li>So is D</li>
                </ul>
            </li>
            <li>E is normal again</li>
        </ol>
        <b>Bold</b><u>Underlined <i>And italic</i></u>.
        <li>This item had no list :(</li>
    </div>
</body>
</html>"""

ascii = """
===============================================================================
Myokit
===============================================================================

Text example 1

This should be visible

More text.

Issue

-------------------------------------------------------------------------------
Software
-------------------------------------------------------------------------------

**This is strong** & this is an ampersand.
-------------------------------------------------------------------------------

  * One
  * Two

...............................................................................
Resources
...............................................................................

This has a
line break.

  1 Numbered A.
  2 *Numbered B is italic.*
  3

    * C is double indented
    * So is D

  4 E is normal again

**Bold** _Underlined *And italic* _ .
* This item had no list :(
""".strip()


class AsciifierTest(unittest.TestCase):
    """
    Tests the method to convert html to ascii.
    """
    def test_asciify(self):

        # Compare line by line
        asc = iter(ascii.splitlines())
        for line1 in myokit.formats.html.html2ascii(html).splitlines():
            # Next line will raise exception if the result is longer than ascii
            line2 = next(asc)
            # Check if lines are equal
            self.assertEqual(line1, line2)

        # Next line will raise exception if ascii is longer than the result
        self.assertRaises(StopIteration, next, asc)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the expression writer for OpenCL.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats
import myokit.formats.opencl


class OpenCLExpressionWriterTest(unittest.TestCase):
    """ Test the OpenCL ewriter class. """

    def test_basic(self):

        # Single and double precision and native maths
        ws = myokit.formats.opencl.OpenCLExpressionWriter()
        wd = myokit.formats.opencl.OpenCLExpressionWriter(
            myokit.DOUBLE_PRECISION)
        wn = myokit.formats.opencl.OpenCLExpressionWriter(native_math=False)

        a = myokit.Name(myokit.Model().add_component('c').add_variable('a'))
        b = myokit.Number('12', 'pF')

        # Name
        self.assertEqual(ws.ex(a), 'c.a')
        self.assertEqual(wd.ex(a), 'c.a')
        self.assertEqual(wn.ex(a), 'c.a')

        # Number with unit
        self.assertEqual(ws.ex(b), '12.0f')
        self.assertEqual(wd.ex(b), '12.0')
        self.assertEqual(wn.ex(b), '12.0f')

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(ws.ex(x), '12.0f')
        self.assertEqual(wd.ex(x), '12.0')
        self.assertEqual(wn.ex(x), '12.0f')
        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertEqual(ws.ex(x), '(-12.0f)')
        self.assertEqual(wd.ex(x), '(-12.0)')
        self.assertEqual(wn.ex(x), '(-12.0f)')

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(ws.ex(x), 'c.a + 12.0f')
        self.assertEqual(wd.ex(x), 'c.a + 12.0')
        self.assertEqual(wn.ex(x), 'c.a + 12.0f')
        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(ws.ex(x), 'c.a - 12.0f')
        self.assertEqual(wd.ex(x), 'c.a - 12.0')
        self.assertEqual(wn.ex(x), 'c.a - 12.0f')
        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(ws.ex(x), 'c.a * 12.0f')
        self.assertEqual(wd.ex(x), 'c.a * 12.0')
        self.assertEqual(wn.ex(x), 'c.a * 12.0f')
        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(ws.ex(x), 'c.a / 12.0f')
        self.assertEqual(wd.ex(x), 'c.a / 12.0')
        self.assertEqual(wn.ex(x), 'c.a / 12.0f')

        # Quotient
        x = myokit.Quotient(a, b)
        self.assertEqual(ws.ex(x), 'floor(c.a / 12.0f)')
        self.assertEqual(wd.ex(x), 'floor(c.a / 12.0)')
        self.assertEqual(wn.ex(x), 'floor(c.a / 12.0f)')
        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(ws.ex(x), 'c.a - 12.0f * (floor(c.a / 12.0f))')
        self.assertEqual(wd.ex(x), 'c.a - 12.0 * (floor(c.a / 12.0))')
        self.assertEqual(wn.ex(x), 'c.a - 12.0f * (floor(c.a / 12.0f))')

    def test_functions(self):

        # Single and double precision and native maths
        ws = myokit.formats.opencl.OpenCLExpressionWriter()
        wd = myokit.formats.opencl.OpenCLExpressionWriter(
            myokit.DOUBLE_PRECISION)
        wn = myokit.formats.opencl.OpenCLExpressionWriter(native_math=False)

        a = myokit.Name(myokit.Model().add_component('c').add_variable('a'))
        b = myokit.Number('12', 'pF')

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(ws.ex(x), 'pow(c.a, 12.0f)')
        self.assertEqual(wd.ex(x), 'pow(c.a, 12.0)')
        self.assertEqual(wn.ex(x), 'pow(c.a, 12.0f)')
        # Square
        x = myokit.Power(a, myokit.Number(2))
        self.assertEqual(ws.ex(x), '(c.a * c.a)')
        self.assertEqual(wd.ex(x), '(c.a * c.a)')
        self.assertEqual(wn.ex(x), '(c.a * c.a)')
        # Square with brackets
        x = myokit.Power(myokit.Plus(a, b), myokit.Number(2))
        self.assertEqual(ws.ex(x), '((c.a + 12.0f) * (c.a + 12.0f))')
        self.assertEqual(wd.ex(x), '((c.a + 12.0) * (c.a + 12.0))')
        self.assertEqual(wn.ex(x), '((c.a + 12.0f) * (c.a + 12.0f))')
        # Sqrt
        x = myokit.Sqrt(b)
        self.assertEqual(ws.ex(x), 'native_sqrt(12.0f)')
        self.assertEqual(wd.ex(x), 'native_sqrt(12.0)')
        self.assertEqual(wn.ex(x), 'sqrt(12.0f)')
        # Exp
        x = myokit.Exp(a)
        self.assertEqual(ws.ex(x), 'native_exp(c.a)')
        self.assertEqual(wd.ex(x), 'native_exp(c.a)')
        self.assertEqual(wn.ex(x), 'exp(c.a)')
        # Log(a)
        x = myokit.Log(b)
        self.assertEqual(ws.ex(x), 'native_log(12.0f)')
        self.assertEqual(wd.ex(x), 'native_log(12.0)')
        self.assertEqual(wn.ex(x), 'log(12.0f)')
        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertEqual(ws.ex(x), '(native_log(c.a) / native_log(12.0f))')
        self.assertEqual(wd.ex(x), '(native_log(c.a) / native_log(12.0))')
        self.assertEqual(wn.ex(x), '(log(c.a) / log(12.0f))')
        # Log10
        x = myokit.Log10(b)
        self.assertEqual(ws.ex(x), 'native_log10(12.0f)')
        self.assertEqual(wd.ex(x), 'native_log10(12.0)')
        self.assertEqual(wn.ex(x), 'log10(12.0f)')

        # Sin
        x = myokit.Sin(b)
        self.assertEqual(ws.ex(x), 'native_sin(12.0f)')
        self.assertEqual(wd.ex(x), 'native_sin(12.0)')
        self.assertEqual(wn.ex(x), 'sin(12.0f)')
        # Cos
        x = myokit.Cos(b)
        self.assertEqual(ws.ex(x), 'native_cos(12.0f)')
        self.assertEqual(wd.ex(x), 'native_cos(12.0)')
        self.assertEqual(wn.ex(x), 'cos(12.0f)')
        # Tan
        x = myokit.Tan(b)
        self.assertEqual(ws.ex(x), 'native_tan(12.0f)')
        self.assertEqual(wd.ex(x), 'native_tan(12.0)')
        self.assertEqual(wn.ex(x), 'tan(12.0f)')
        # ASin
        x = myokit.ASin(b)
        self.assertEqual(ws.ex(x), 'asin(12.0f)')
        self.assertEqual(wd.ex(x), 'asin(12.0)')
        self.assertEqual(wn.ex(x), 'asin(12.0f)')
        # ACos
        x = myokit.ACos(b)
        self.assertEqual(ws.ex(x), 'acos(12.0f)')
        self.assertEqual(wd.ex(x), 'acos(12.0)')
        self.assertEqual(wn.ex(x), 'acos(12.0f)')
        # ATan
        x = myokit.ATan(b)
        self.assertEqual(ws.ex(x), 'atan(12.0f)')
        self.assertEqual(wd.ex(x), 'atan(12.0)')
        self.assertEqual(wn.ex(x), 'atan(12.0f)')

        # Floor
        x = myokit.Floor(b)
        self.assertEqual(ws.ex(x), 'floor(12.0f)')
        self.assertEqual(wd.ex(x), 'floor(12.0)')
        self.assertEqual(wn.ex(x), 'floor(12.0f)')
        # Ceil
        x = myokit.Ceil(b)
        self.assertEqual(ws.ex(x), 'ceil(12.0f)')
        self.assertEqual(wd.ex(x), 'ceil(12.0)')
        self.assertEqual(wn.ex(x), 'ceil(12.0f)')
        # Abs
        x = myokit.Abs(b)
        self.assertEqual(ws.ex(x), 'fabs(12.0f)')
        self.assertEqual(wd.ex(x), 'fabs(12.0)')
        self.assertEqual(wn.ex(x), 'fabs(12.0f)')

    def test_conditional(self):

        # Single and double precision and native maths
        ws = myokit.formats.opencl.OpenCLExpressionWriter()
        wd = myokit.formats.opencl.OpenCLExpressionWriter(
            myokit.DOUBLE_PRECISION)
        wn = myokit.formats.opencl.OpenCLExpressionWriter(native_math=False)

        a = myokit.Name(myokit.Model().add_component('c').add_variable('a'))
        b = myokit.Number('12', 'pF')
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')

        # Equal
        x = myokit.Equal(a, b)
        self.assertEqual(ws.ex(x), '(c.a == 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a == 12.0)')
        self.assertEqual(wn.ex(x), '(c.a == 12.0f)')
        x = myokit.Equal(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) == (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) == (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) == (2.0f < 1.0f))')
        x = myokit.Equal(cond1, b)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) == (12.0f != 0.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) == (12.0 != 0.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) == (12.0f != 0.0f))')
        x = myokit.Equal(a, cond2)
        self.assertEqual(ws.ex(x), '((c.a != 0.0f) == (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((c.a != 0.0) == (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((c.a != 0.0f) == (2.0f < 1.0f))')

        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertEqual(ws.ex(x), '(c.a != 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a != 12.0)')
        self.assertEqual(wn.ex(x), '(c.a != 12.0f)')
        x = myokit.NotEqual(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) != (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) != (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) != (2.0f < 1.0f))')
        x = myokit.NotEqual(cond1, b)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) != (12.0f != 0.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) != (12.0 != 0.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) != (12.0f != 0.0f))')
        x = myokit.NotEqual(a, cond2)
        self.assertEqual(ws.ex(x), '((c.a != 0.0f) != (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((c.a != 0.0) != (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((c.a != 0.0f) != (2.0f < 1.0f))')

        # More
        x = myokit.More(a, b)
        self.assertEqual(ws.ex(x), '(c.a > 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a > 12.0)')
        self.assertEqual(wn.ex(x), '(c.a > 12.0f)')
        x = myokit.More(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) > (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) > (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) > (2.0f < 1.0f))')
        x = myokit.More(cond1, b)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) > (12.0f != 0.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) > (12.0 != 0.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) > (12.0f != 0.0f))')
        x = myokit.More(a, cond2)
        self.assertEqual(ws.ex(x), '((c.a != 0.0f) > (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((c.a != 0.0) > (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((c.a != 0.0f) > (2.0f < 1.0f))')

        # Less
        x = myokit.Less(a, b)
        self.assertEqual(ws.ex(x), '(c.a < 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a < 12.0)')
        self.assertEqual(wn.ex(x), '(c.a < 12.0f)')
        x = myokit.Less(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) < (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) < (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) < (2.0f < 1.0f))')
        x = myokit.Less(cond1, b)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) < (12.0f != 0.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) < (12.0 != 0.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) < (12.0f != 0.0f))')
        x = myokit.Less(a, cond2)
        self.assertEqual(ws.ex(x), '((c.a != 0.0f) < (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((c.a != 0.0) < (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((c.a != 0.0f) < (2.0f < 1.0f))')

        # MoreEqual
        x = myokit.MoreEqual(a, b)
        self.assertEqual(ws.ex(x), '(c.a >= 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a >= 12.0)')
        self.assertEqual(wn.ex(x), '(c.a >= 12.0f)')
        x = myokit.MoreEqual(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) >= (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) >= (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) >= (2.0f < 1.0f))')
        x = myokit.MoreEqual(cond1, b)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) >= (12.0f != 0.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) >= (12.0 != 0.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) >= (12.0f != 0.0f))')
        x = myokit.MoreEqual(a, cond2)
        self.assertEqual(ws.ex(x), '((c.a != 0.0f) >= (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((c.a != 0.0) >= (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((c.a != 0.0f) >= (2.0f < 1.0f))')

        # LessEqual
        x = myokit.LessEqual(a, b)
        self.assertEqual(ws.ex(x), '(c.a <= 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a <= 12.0)')
        self.assertEqual(wn.ex(x), '(c.a <= 12.0f)')
        x = myokit.LessEqual(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) <= (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) <= (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) <= (2.0f < 1.0f))')
        x = myokit.LessEqual(cond1, b)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) <= (12.0f != 0.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) <= (12.0 != 0.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) <= (12.0f != 0.0f))')
        x = myokit.LessEqual(a, cond2)
        self.assertEqual(ws.ex(x), '((c.a != 0.0f) <= (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((c.a != 0.0) <= (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((c.a != 0.0f) <= (2.0f < 1.0f))')

    def test_logical(self):

        # Single and double precision and native maths
        ws = myokit.formats.opencl.OpenCLExpressionWriter()
        wd = myokit.formats.opencl.OpenCLExpressionWriter(
            myokit.DOUBLE_PRECISION)
        wn = myokit.formats.opencl.OpenCLExpressionWriter(native_math=False)

        a = myokit.Name(myokit.Model().add_component('c').add_variable('a'))
        b = myokit.Number('12', 'pF')
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        condx = myokit.Number(1.2)

        # Not
        x = myokit.Not(cond1)
        self.assertEqual(ws.ex(x), '!((5.0f > 3.0f))')
        self.assertEqual(wd.ex(x), '!((5.0 > 3.0))')
        self.assertEqual(wn.ex(x), '!((5.0f > 3.0f))')
        x = myokit.Not(condx)
        self.assertEqual(ws.ex(x), '!((1.2f != 0.0f))')
        self.assertEqual(wd.ex(x), '!((1.2 != 0.0))')
        self.assertEqual(wn.ex(x), '!((1.2f != 0.0f))')

        # And
        x = myokit.And(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) && (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) && (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) && (2.0f < 1.0f))')
        x = myokit.And(condx, cond2)
        self.assertEqual(ws.ex(x), '((1.2f != 0.0f) && (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((1.2 != 0.0) && (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((1.2f != 0.0f) && (2.0f < 1.0f))')
        x = myokit.And(cond1, condx)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) && (1.2f != 0.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) && (1.2 != 0.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) && (1.2f != 0.0f))')

        # Or
        x = myokit.Or(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) || (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) || (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) || (2.0f < 1.0f))')
        x = myokit.Or(condx, cond2)
        self.assertEqual(ws.ex(x), '((1.2f != 0.0f) || (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((1.2 != 0.0) || (2.0 < 1.0))')
        self.assertEqual(wn.ex(x), '((1.2f != 0.0f) || (2.0f < 1.0f))')
        x = myokit.Or(cond1, condx)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) || (1.2f != 0.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) || (1.2 != 0.0))')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) || (1.2f != 0.0f))')

        # If
        x = myokit.If(cond1, a, b)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) ? c.a : 12.0f)')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) ? c.a : 12.0)')
        self.assertEqual(wn.ex(x), '((5.0f > 3.0f) ? c.a : 12.0f)')
        x = myokit.If(condx, a, b)
        self.assertEqual(ws.ex(x), '((1.2f != 0.0f) ? c.a : 12.0f)')
        self.assertEqual(wd.ex(x), '((1.2 != 0.0) ? c.a : 12.0)')
        self.assertEqual(wn.ex(x), '((1.2f != 0.0f) ? c.a : 12.0f)')

        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            ws.ex(x),
            '((5.0f > 3.0f) ? c.a : ((2.0f < 1.0f) ? 12.0f : 1.0f))')
        self.assertEqual(
            wd.ex(x),
            '((5.0 > 3.0) ? c.a : ((2.0 < 1.0) ? 12.0 : 1.0))')
        self.assertEqual(
            wn.ex(x),
            '((5.0f > 3.0f) ? c.a : ((2.0f < 1.0f) ? 12.0f : 1.0f))')
        x = myokit.Piecewise(condx, a, condx, b, c)
        self.assertEqual(
            ws.ex(x),
            '((1.2f != 0.0f) ? c.a : ((1.2f != 0.0f) ? 12.0f : 1.0f))')
        self.assertEqual(
            wd.ex(x),
            '((1.2 != 0.0) ? c.a : ((1.2 != 0.0) ? 12.0 : 1.0))')
        self.assertEqual(
            wn.ex(x),
            '((1.2f != 0.0f) ? c.a : ((1.2f != 0.0f) ? 12.0f : 1.0f))')

    def test_fetching(self):
        # Test fetching using ewriter method
        w = myokit.formats.ewriter('opencl')
        self.assertIsInstance(w, myokit.formats.opencl.OpenCLExpressionWriter)

    def test_bad_operand(self):
        # Test without a Myokit expression
        w = myokit.formats.opencl.OpenCLExpressionWriter()
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', w.ex, 7)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests protocol creation via myokit.pacing
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit


class PacingFactoryTest(unittest.TestCase):
    """
    Tests the pacing factory module `myokit.pacing`.
    """

    def test_blocktrain(self):
        # Test creation of a block train protocol.

        period, duration, offset, level, limit = 1000, 1, 0, 1, 0
        p = myokit.pacing.blocktrain(period, duration, offset, level, limit)
        e = p.events()
        self.assertEqual(len(e), 1)
        e = e[0]
        self.assertEqual(e.level(), level)
        self.assertEqual(e.start(), offset)
        self.assertEqual(e.duration(), duration)
        self.assertEqual(e.period(), period)
        self.assertEqual(e.multiplier(), limit)

        period, duration, offset, level, limit = 500, 0.5, 79, 2.1, 14
        p = myokit.pacing.blocktrain(period, duration, offset, level, limit)
        e = p.events()
        self.assertEqual(len(e), 1)
        e = e[0]
        self.assertEqual(e.level(), level)
        self.assertEqual(e.start(), offset)
        self.assertEqual(e.duration(), duration)
        self.assertEqual(e.period(), period)
        self.assertEqual(e.multiplier(), limit)

    def test_bpm2bcl(self):
        # Test conversion from beats-per-minute to cycle length.

        # Milliseconds
        self.assertEqual(myokit.pacing.bpm2bcl(60), 1000)
        self.assertEqual(myokit.pacing.bpm2bcl(30), 2000)
        self.assertEqual(myokit.pacing.bpm2bcl(120), 500)

        # Seconds
        self.assertEqual(myokit.pacing.bpm2bcl(60, 1), 1)
        self.assertEqual(myokit.pacing.bpm2bcl(30, 1), 2)
        self.assertEqual(myokit.pacing.bpm2bcl(120, 1), 0.5)

    def test_constant(self):
        # Test the creation of a constant protocol.

        level = 0.5
        p = myokit.pacing.constant(level)
        s = myokit.PacingSystem(p)
        self.assertEqual(len(p.events()), 1)
        self.assertEqual(s.pace(), level)
        s.advance(100)
        self.assertEqual(s.pace(), level)
        for i in range(10):
            s.advance(s.next_time())
            self.assertEqual(s.pace(), level)

    def test_steptrain(self):
        # Test the creation of a step protocol.

        vs = [-100, -80, 40, -20]
        vhold = -80
        tpre = 200
        tstep = 1000
        tpost = 800

        p = myokit.pacing.steptrain(vs, vhold, tpre, tstep, tpost)
        self.assertEqual(len(p.events()), 3 * len(vs))
        self.assertEqual(
            p.characteristic_time(), len(vs) * (tpre + tstep + tpost))
        p = iter(p)
        t = 0
        for v in vs:
            e = next(p)
            self.assertEqual(e.level(), vhold)
            self.assertEqual(e.start(), t)
            self.assertEqual(e.duration(), tpre)
            self.assertEqual(e.period(), 0)
            self.assertEqual(e.multiplier(), 0)
            t += tpre

            e = next(p)
            self.assertEqual(e.level(), v)
            self.assertEqual(e.start(), t)
            self.assertEqual(e.duration(), tstep)
            self.assertEqual(e.period(), 0)
            self.assertEqual(e.multiplier(), 0)
            t += tstep

            e = next(p)
            self.assertEqual(e.level(), vhold)
            self.assertEqual(e.start(), t)
            self.assertEqual(e.duration(), tpost)
            self.assertEqual(e.period(), 0)
            self.assertEqual(e.multiplier(), 0)
            t += tpost

    def test_steptrain_linear(self):
        # Test the creation of a step protocol with linear steps.

        # Incrementing steps
        vs = [-40, -20, 0, 20]
        vmin = -40
        vmax = 40
        dv = 20
        vhold = -80
        tpre = 200
        tstep = 1000
        tpost = 800

        p = myokit.pacing.steptrain_linear(
            vmin, vmax, dv, vhold, tpre, tstep, tpost)
        self.assertEqual(len(p.events()), 3 * len(vs))
        self.assertEqual(
            p.characteristic_time(), len(vs) * (tpre + tstep + tpost))
        p = iter(p)
        t = 0
        for v in vs:
            e = next(p)
            self.assertEqual(e.level(), vhold)
            self.assertEqual(e.start(), t)
            self.assertEqual(e.duration(), tpre)
            self.assertEqual(e.period(), 0)
            self.assertEqual(e.multiplier(), 0)
            t += tpre

            e = next(p)
            self.assertEqual(e.level(), v)
            self.assertEqual(e.start(), t)
            self.assertEqual(e.duration(), tstep)
            self.assertEqual(e.period(), 0)
            self.assertEqual(e.multiplier(), 0)
            t += tstep

            e = next(p)
            self.assertEqual(e.level(), vhold)
            self.assertEqual(e.start(), t)
            self.assertEqual(e.duration(), tpost)
            self.assertEqual(e.period(), 0)
            self.assertEqual(e.multiplier(), 0)
            t += tpost

        # Decrementing steps
        vs = [40, 20, 0, -20]
        vmin = 40
        vmax = -40
        dv = -20
        vhold = -80
        tpre = 200
        tstep = 1000
        tpost = 800

        p = myokit.pacing.steptrain_linear(
            vmin, vmax, dv, vhold, tpre, tstep, tpost)
        self.assertEqual(len(p.events()), 3 * len(vs))
        self.assertEqual(
            p.characteristic_time(), len(vs) * (tpre + tstep + tpost))
        p = iter(p)
        t = 0
        for v in vs:
            e = next(p)
            self.assertEqual(e.level(), vhold)
            self.assertEqual(e.start(), t)
            self.assertEqual(e.duration(), tpre)
            self.assertEqual(e.period(), 0)
            self.assertEqual(e.multiplier(), 0)
            t += tpre

            e = next(p)
            self.assertEqual(e.level(), v)
            self.assertEqual(e.start(), t)
            self.assertEqual(e.duration(), tstep)
            self.assertEqual(e.period(), 0)
            self.assertEqual(e.multiplier(), 0)
            t += tstep

            e = next(p)
            self.assertEqual(e.level(), vhold)
            self.assertEqual(e.start(), t)
            self.assertEqual(e.duration(), tpost)
            self.assertEqual(e.period(), 0)
            self.assertEqual(e.multiplier(), 0)
            t += tpost

    def test_steptrain_bad_values(self):
        # Test the creation of a step protocol with illegal times.

        vs = [-100, -80, 40, -20]
        vhold = -80
        tpre = 200
        tstep = 1000
        tpost = 800

        myokit.pacing.steptrain(vs, vhold, tpre, tstep, tpost)
        self.assertRaises(
            ValueError, myokit.pacing.steptrain, vs, vhold, -1, tstep, tpost)
        self.assertRaises(
            ValueError, myokit.pacing.steptrain, vs, vhold, tpre, -1, tpost)
        self.assertRaises(
            ValueError, myokit.pacing.steptrain, vs, vhold, tpre, tstep, -1)

    def test_steptrain_linear_bad_values(self):
        # Test the creation of a step protocol with linear steps and bad
        # values.

        # Incrementing steps
        vmin = -40
        vmax = 40
        dv = 20
        vhold = -80
        tpre = 200
        tstep = 1000
        tpost = 800

        # Wrong direction of dv
        myokit.pacing.steptrain_linear(
            vmin, vmax, dv, vhold, tpre, tstep, tpost)
        self.assertRaises(
            ValueError, myokit.pacing.steptrain_linear, vmin, vmax, -dv, vhold,
            tpre, tstep, tpost)
        myokit.pacing.steptrain_linear(
            vmax, vmin, -dv, vhold, tpre, tstep, tpost)
        self.assertRaises(
            ValueError, myokit.pacing.steptrain_linear, vmax, vmin, dv, vhold,
            tpre, tstep, tpost)

        # Bad times
        self.assertRaises(
            ValueError, myokit.pacing.steptrain_linear, vmin, vmax, dv, vhold,
            -1, tstep, tpost)
        self.assertRaises(
            ValueError, myokit.pacing.steptrain_linear, vmin, vmax, dv, vhold,
            tpre, -1, tpost)
        self.assertRaises(
            ValueError, myokit.pacing.steptrain_linear, vmin, vmax, dv, vhold,
            tpre, tstep, -1)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the Quantity class
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit


class QuantityTest(unittest.TestCase):
    """
    Tests the Quantity class for unit arithmetic.
    """

    def test_as_rhs(self):
        # Test Quantity use in set_rhs.
        from myokit import Quantity as Q

        m = myokit.Model()
        c = m.add_component('a')
        v = c.add_variable('v')
        a = Q('10 [mV]')
        v.set_rhs(a)
        self.assertEqual(v.rhs().unit(), myokit.units.mV)
        self.assertEqual(v.eval(), 10)

    def test_cast(self):
        # Test :meth:`Quanity.cast()`.

        from myokit import Quantity as Q

        a = Q('10 [uA]')
        b = a.cast('mV')
        self.assertEqual(a, Q('10 [uA]'))
        self.assertEqual(b, Q('10 [mV]'))

    def test_convert(self):
        # Test :meth:`Quantity.convert()`.

        from myokit import Quantity as Q

        a = Q('10 [mV]')
        self.assertEqual(a.convert('V'), Q('0.01 [V]'))

    def test_creation_and_str(self):
        # Test Quanity creation and :meth:`Quantity.__str__()`.

        from myokit import Quantity as Q

        # Creation and string representation
        a = Q('10 [mV]')
        self.assertEqual(float(a), 10)
        self.assertEqual(str(a), '10.0 [mV]')
        a = Q('2', myokit.units.uA)
        self.assertEqual(float(a), 2)
        self.assertEqual(str(a), '2.0 [uA]')
        a = Q(-2, 'uA/cm^2')
        self.assertEqual(float(a), -2)
        self.assertEqual(str(a), '-2.0 [uA/cm^2]')
        a = Q(3.0e1, 'km/N/g')
        self.assertEqual(float(a), 30)
        self.assertEqual(str(a), '30.0 [s^2/g^2]')
        a = Q(4)
        self.assertEqual(float(a), 4)
        self.assertEqual(str(a), '4.0 [1]')

        # Repr
        self.assertEqual(repr(a), '4.0 [1]')

        # Creation from a myokit Number
        x = myokit.Number(2)
        a = Q(x)
        self.assertRaisesRegex(
            ValueError, 'Cannot specify', Q, x, myokit.Unit())

        # Creation with string value
        x = myokit.Quantity('2')
        x = myokit.Quantity('2 [mV]')
        self.assertRaisesRegex(
            ValueError, 'could not be converted', myokit.Quantity, int)
        self.assertRaisesRegex(
            ValueError, 'Failed to parse', myokit.Quantity, 'wolf')
        self.assertRaisesRegex(
            ValueError, 'Failed to parse', myokit.Quantity, 'a [mV]')
        x = myokit.Quantity('2', 'mV')
        self.assertRaisesRegex(
            ValueError, 'Two units', myokit.Quantity, '2 [mV]', 'mV')

    def test_eq(self):
        # Test :meth:`Quantity.__eq__()`.
        from myokit import Quantity as Q

        a = Q('10 [mV]')
        self.assertEqual(a, Q('10 [mV]'))
        self.assertNotEqual(a, Q('11 [mV]'))
        self.assertNotEqual(a, Q('10 [mA]'))
        self.assertNotEqual(a, Q('10 [V]'))
        self.assertNotEqual(a, Q('0.01 [V]'))
        self.assertEqual(a, Q('0.01 [V]').convert('mV'))
        self.assertFalse(Q(4) == 4)
        self.assertFalse(4 == Q(4))

    def test_hash(self):
        # Test has does not change in quantity's lifetime

        try:
            u = myokit.units.m**8
            q1 = myokit.Quantity(1, u)
            h1 = hash(q1)
            myokit.Unit.register_preferred_representation('abc', u)
            q2 = myokit.Quantity(1, u)
            h2 = hash(q2)
            self.assertEqual(h1, h2)
        finally:
            # Bypassing the public API, this is bad test design!
            if u in myokit.Unit._preferred_representations:
                del myokit.Unit._preferred_representations[u]

    def test_number_conversion(self):
        # Test Quantity conversion from and to number.

        from myokit import Quantity as Q
        from myokit import Number as N

        # Conversion from number
        d = N(4)
        self.assertIsNone(d.unit())
        e = Q(d)
        self.assertEqual(float(e), 4)
        self.assertIsNotNone(e.unit())
        self.assertEqual(e.unit(), myokit.units.dimensionless)
        self.assertEqual(str(e), '4.0 [1]')

        # Conversion to number
        a = Q('10 [mV]')
        b = myokit.Number(a)
        self.assertEqual(b.eval(), 10)
        self.assertEqual(b.unit(), myokit.units.mV)

    def test_operators(self):
        # Test overloaded operators for Quantity.

        from myokit import Quantity as Q

        # Addition
        a = Q('10 [mV]')
        b = Q('3 [mV]')
        self.assertEqual(a + b, Q('13 [mV]'))
        self.assertEqual(a + Q('-4.2 [mV]'), Q('5.8 [mV]'))

        def add(a, b):
            return a + b
        self.assertEqual(add(a, b), Q('13 [mV]'))
        b = Q('3 [V]')
        self.assertRaises(myokit.IncompatibleUnitError, add, a, b)
        self.assertRaises(myokit.IncompatibleUnitError, add, a, 3)
        a = Q(4)
        self.assertEqual(a + 2, Q(6))
        self.assertEqual(a + 3, Q(7, myokit.units.dimensionless))
        self.assertEqual(a + 4, Q('8'))
        self.assertEqual(a + 5, Q('9 [1]'))
        self.assertEqual(a + 2, 2 + a)
        self.assertRaises(myokit.IncompatibleUnitError, add, a, b)

        # Subtraction
        a = Q('10 [mV]')
        b = Q('3 [mV]')
        self.assertEqual(a - b, Q('7 [mV]'))
        self.assertEqual(a - Q('-4.2 [mV]'), Q('14.2 [mV]'))

        def sub(a, b):
            return a - b
        self.assertEqual(sub(a, b), Q('7 [mV]'))
        b = Q('3 [V]')
        self.assertRaises(myokit.IncompatibleUnitError, sub, a, b)
        a = Q(10)
        self.assertEqual(a - 2, Q(8))
        self.assertEqual(a - 3, Q(7, myokit.units.dimensionless))
        self.assertEqual(a - 4, Q('6'))
        self.assertEqual(a - 5, Q('5 [1]'))
        self.assertEqual(Q(10) - 2, 10 - Q(2))
        self.assertRaises(myokit.IncompatibleUnitError, sub, a, b)

        # Multiplication
        a = Q(2)
        b = Q(3)
        self.assertEqual(a * b, Q(6))
        self.assertEqual(3 * b, Q(9))
        self.assertEqual(a * 4, Q(8))
        a = Q('10 [mV]')
        b = Q('2 [uA]')
        self.assertEqual(a * b, Q('20 [nW]'))
        self.assertEqual(a * 5, 5 * a)
        self.assertEqual(a * 5, Q('50 [mV]'))
        self.assertEqual(a * 1000, Q('10000 [mV]'))
        self.assertEqual((a * 1000).convert('V'), Q('10 [V]'))

        # Division
        a = Q('10 [uA]')
        b = Q('2 [mV]')
        c = a / b
        self.assertEqual(c, Q('5 [mS]'))
        self.assertEqual(float(c), 5.0)
        self.assertEqual(str(c), '5.0 [mS]')
        self.assertEqual(b.convert('V'), Q('0.002 [V]'))
        self.assertRaises(myokit.IncompatibleUnitError, a.convert, 'V')
        a = Q(3)
        self.assertEqual(a / 2, Q(3 / 2))

        a = Q('10 [uA]')
        self.assertEqual(a / 2, Q('5 [uA]'))
        self.assertEqual(2 / a, Q('0.2 [1/uA]'))

        # Powers
        a = Q('10 [uA]')
        self.assertEqual(a ** 2, Q('100 [uA^2]'))
        self.assertEqual(a ** 3, Q('1000 [uA^3]'))
        self.assertEqual(a ** Q(3), Q('1000 [uA^3]'))
        a = Q('10 [uA]') ** 3
        b = a ** (1 / 3)
        self.assertEqual(b.unit(), myokit.units.uA)
        self.assertAlmostEqual(b.value(), 10)
        with self.assertRaises(myokit.IncompatibleUnitError):
            a ** a


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the CellML 1.0/1.1 API.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats.cellml.v1 as cellml

from myokit.tests import WarningCollector


class TestCellML1AnnotatedElement(unittest.TestCase):
    """ Tests for cellml.AnnotatableElement. """

    def test_cmeta_id(self):
        # Tests the cmeta_id methods.

        m = cellml.Model('m')
        c = m.add_component('c')

        # Test getting and setting
        self.assertIsNone(m.cmeta_id())
        self.assertIsNone(c.cmeta_id())
        m_id = 'this-is-m'
        c_id = 'this-is-c'
        m.set_cmeta_id(m_id)
        c.set_cmeta_id(c_id)
        self.assertEqual(m.cmeta_id(), m_id)
        self.assertEqual(c.cmeta_id(), c_id)

        # Test getting by cmeta id
        self.assertEqual(m.element_with_cmeta_id(m_id), m)
        self.assertEqual(m.element_with_cmeta_id(c_id), c)

        # Test unsetting
        m.set_cmeta_id(None)
        c.set_cmeta_id(None)
        self.assertIsNone(m.cmeta_id())
        self.assertIsNone(c.cmeta_id())
        self.assertRaises(KeyError, m.element_with_cmeta_id, m_id)
        self.assertRaises(KeyError, m.element_with_cmeta_id, c_id)

        # Test multiple setting
        c.set_cmeta_id(c_id)
        c.set_cmeta_id(c_id)
        c.set_cmeta_id(c_id)
        self.assertEqual(c.cmeta_id(), c_id)
        self.assertEqual(m.element_with_cmeta_id(c_id), c)
        c.set_cmeta_id(None)
        self.assertIsNone(c.cmeta_id())
        self.assertRaises(KeyError, m.element_with_cmeta_id, c_id)

        # Test id is checked for well-formedness
        c.set_cmeta_id('123')
        self.assertRaisesRegex(
            cellml.CellMLError, 'empty string', c.set_cmeta_id, '   ')

        # Test bad set call doesn't mess up state
        self.assertEqual(c.cmeta_id(), '123')
        self.assertEqual(m.element_with_cmeta_id('123'), c)
        c.set_cmeta_id(c_id)
        self.assertRaises(KeyError, m.element_with_cmeta_id, '123')

        # Test duplicates
        m.set_cmeta_id(m_id)
        self.assertRaisesRegex(
            cellml.CellMLError, 'unique', c.set_cmeta_id, m_id)


class TestCellML1Component(unittest.TestCase):
    """ Tests for ``cellml.v1.Component``. """

    def test_creation(self):
        # Test component creation

        m = cellml.Model('m')
        c = m.add_component('c')

        self.assertRaisesRegex(
            cellml.CellMLError, 'valid CellML identifier',
            m.add_component, '123')

    def test_add_find_units(self):
        # Tests adding and finding units

        m = cellml.Model('m')
        c = m.add_component('c')

        # Test getting non-existent
        self.assertRaisesRegex(
            cellml.CellMLError, 'Unknown', c.find_units, 'wooster')

        # Test adding
        u = c.add_units('wooster', myokit.units.lumen)
        self.assertEqual(c.find_units('wooster'), u)

        # Test doubles are not allowed
        self.assertRaisesRegex(
            cellml.CellMLError, 'same name', c.add_units,
            'wooster', myokit.units.meter)

        # But shadowing is
        v = m.add_units('wooster', myokit.units.meter)
        self.assertEqual(c.find_units('wooster'), u)
        self.assertEqual(m.find_units('wooster'), v)

        # Test iteration
        w = c.add_units('jarvis', myokit.units.meter)
        us = [unit for unit in c.units()]
        self.assertEqual(len(us), 2)
        self.assertIn(u, us)
        self.assertIn(w, us)

        # Test finding names (finds the last added with that unit)
        self.assertEqual(c.find_units_name(myokit.units.meter), 'jarvis')
        self.assertEqual(c.find_units_name(myokit.units.volt), 'volt')
        self.assertRaisesRegex(
            cellml.CellMLError, 'No name found for myokit unit',
            c.find_units_name, 1 / myokit.units.volt)

    def test_add_get_variable(self):
        # Tests adding and getting variables

        # Test adding
        m = cellml.Model('m')
        c = m.add_component('c')
        self.assertEqual(len(c), 0)
        v = c.add_variable('v', 'volt')
        self.assertEqual(len(c), 1)

        # Test getting
        self.assertIs(c.variable('v'), v)

        # Test adding duplicate
        self.assertRaisesRegex(
            cellml.CellMLError, 'unique', c.add_variable, 'v', 'meter')

        # Test iteration over all
        w = c.add_variable('w', 'volt')
        vs = [x for x in c.variables()]
        self.assertEqual(len(vs), 2)
        self.assertIn(v, vs)
        self.assertIn(w, vs)

        # Test creating a variable with celsius
        c = cellml.Model('m').add_component('c')
        self.assertRaises(
            cellml.UnsupportedBaseUnitsError,
            c.add_variable, 'v', 'celsius')

        # Rest of creation is tested in variable test

    def test_model(self):
        # Tests Component.model()

        m = cellml.Model('m')
        c = m.add_component('pixie')
        self.assertIs(c.model(), m)

    def test_name(self):
        # Tests Component.name()

        m = cellml.Model('m')
        c = m.add_component('trixie')
        self.assertIs(c.name(), 'trixie')

    def test_sequence_interface(self):
        # Tests the sequence interface on a component

        m = cellml.Model('m')
        a = m.add_component('Stacey')
        b = m.add_component('Jane')
        c = m.add_component('Mary_jo_lisa')
        self.assertIs(a, m['Stacey'])
        self.assertIs(b, m['Jane'])
        self.assertIs(c, m['Mary_jo_lisa'])
        self.assertIn('Stacey', m)
        self.assertIn('Jane', m)
        self.assertIn('Mary_jo_lisa', m)
        self.assertEqual(len(m), 3)

    def test_set_get_parent(self):
        # Tests setting and getting of parents

        # Test getting
        m = cellml.Model('m')
        a = m.add_component('Stacey')
        b = m.add_component('Jane')
        c = m.add_component('Mary')
        d = m.add_component('Jo')
        e = m.add_component('Lisa')
        self.assertIsNone(a.parent())
        self.assertIsNone(b.parent())
        self.assertIsNone(c.parent())
        self.assertIsNone(d.parent())
        self.assertIsNone(e.parent())
        self.assertFalse(a.has_children())
        self.assertFalse(b.has_children())
        self.assertFalse(c.has_children())
        self.assertFalse(d.has_children())
        self.assertFalse(e.has_children())
        self.assertEqual(len(list(a.children())), 0)
        self.assertEqual(len(list(b.children())), 0)
        self.assertEqual(len(list(c.children())), 0)
        self.assertEqual(len(list(e.children())), 0)
        self.assertEqual(len(list(e.children())), 0)

        # Test setting
        b.set_parent(a)
        c.set_parent(b)
        d.set_parent(c)
        e.set_parent(c)
        self.assertIsNone(a.parent())
        self.assertIs(b.parent(), a)
        self.assertIs(c.parent(), b)
        self.assertIs(d.parent(), c)
        self.assertIs(e.parent(), c)
        self.assertTrue(a.has_children())
        self.assertTrue(b.has_children())
        self.assertTrue(c.has_children())
        self.assertFalse(d.has_children())
        self.assertFalse(e.has_children())
        self.assertEqual(len(list(a.children())), 1)
        self.assertEqual(len(list(b.children())), 1)
        self.assertEqual(len(list(c.children())), 2)
        self.assertEqual(len(list(d.children())), 0)
        self.assertEqual(len(list(e.children())), 0)
        self.assertIn(b, list(a.children()))
        self.assertIn(c, list(b.children()))
        self.assertIn(d, list(c.children()))
        self.assertIn(e, list(c.children()))

        # Test bad parent
        self.assertRaisesRegex(ValueError, 'v1.Component', b.set_parent, m)
        m2 = cellml.Model('m2')
        c2 = m2.add_component('c2')
        self.assertRaisesRegex(ValueError, 'same model', b.set_parent, c2)

    def test_string_conversion(self):
        # Tests Component.__str__

        m = cellml.Model('mm')
        c = m.add_component('xx')
        self.assertEqual(str(c), 'Component[@name="xx"]')

    def test_validation(self):
        # Tests component validation

        # Test cyclical encapsulation detection
        m = cellml.Model('m')
        a = m.add_component('Stacey')
        b = m.add_component('Jane')
        c = m.add_component('Mary_jo_lisa')
        a.set_parent(b)
        b.set_parent(c)
        c.set_parent(a)
        self.assertRaisesRegex(cellml.CellMLError, 'circular', m.validate)
        c.set_parent(None)
        m.validate()


class TestCellML1Model(unittest.TestCase):
    """ Tests for ``cellml.Model``. """

    def test_add_find_units(self):
        # Tests adding and finding units

        m = cellml.Model('m')

        # Test getting non-existent
        self.assertRaisesRegex(
            cellml.CellMLError, 'Unknown', m.find_units, 'wooster')

        # Test adding
        u = m.add_units('wooster', myokit.units.lumen)
        self.assertEqual(m.find_units('wooster'), u)

        # Test doubles are not allowed
        self.assertRaisesRegex(
            cellml.CellMLError, 'same name',
            m.add_units, 'wooster', myokit.units.meter)

        # Test iteration
        w = m.add_units('jarvis', myokit.units.meter)
        us = [unit for unit in m.units()]
        self.assertEqual(len(us), 2)
        self.assertIn(u, us)
        self.assertIn(w, us)

        # Test finding names (finds the last added with that unit)
        self.assertEqual(m.find_units_name(myokit.units.meter), 'jarvis')
        self.assertEqual(m.find_units_name(myokit.units.volt), 'volt')
        self.assertRaisesRegex(
            cellml.CellMLError, 'No name found for myokit unit',
            m.find_units_name, 1 / myokit.units.volt)

    def test_add_get_component(self):
        # Tests adding and getting components

        # Test adding
        m = cellml.Model('m')
        self.assertEqual(len(m), 0)
        c = m.add_component('c')
        self.assertEqual(len(m), 1)

        # Test getting
        self.assertIs(m.component('c'), c)

        # Test adding duplicate
        self.assertRaisesRegex(
            cellml.CellMLError, 'unique', m.add_component, 'c')

        # Test iteration over all
        d = m.add_component('d')
        cs = [x for x in m.components()]
        self.assertEqual(len(cs), 2)
        self.assertIn(c, cs)
        self.assertIn(d, cs)

        # Rest of creation is tested in component test

    def test_add_connection(self):
        # Tests adding connections

        # Test setting connections and using Variable.value_source()
        m = cellml.Model('m')
        a = m.add_component('a')
        b = m.add_component('b')
        c = m.add_component('c')
        c.set_parent(b)

        x = a.add_variable('x', 'volt', 'out', 'none')
        y = b.add_variable('y', 'volt', 'in', 'out')
        z = c.add_variable('z', 'volt', 'in', 'none')

        m.add_connection(x, y)
        m.add_connection(z, y)
        self.assertIsNone(x.source())
        self.assertIs(y.source(), x)
        self.assertIs(z.source(), y)
        self.assertIs(x.value_source(), x)
        self.assertIs(y.value_source(), x)
        self.assertIs(z.value_source(), x)

        # Test bad connections
        x2 = x
        m = cellml.Model('m')
        a = m.add_component('a')
        b = m.add_component('b')
        c = m.add_component('c')
        c.set_parent(b)
        x = a.add_variable('x', 'volt', 'out', 'none')
        y = b.add_variable('y', 'volt', 'in', 'out')
        z = c.add_variable('z', 'volt', 'in', 'none')

        # Not a variable or wrong model
        self.assertRaisesRegex(
            ValueError, 'variable_1 must be a cellml.v1.Variable.',
            m.add_connection, 'x', y)
        self.assertRaisesRegex(
            ValueError, 'variable_2 must be a cellml.v1.Variable.',
            m.add_connection, y, 'x')
        self.assertRaisesRegex(
            ValueError, 'variable_1 must be a variable from',
            m.add_connection, x2, y)
        self.assertRaisesRegex(
            ValueError, 'variable_2 must be a variable from',
            m.add_connection, y, x2)

        # Connected to self
        self.assertRaisesRegex(
            cellml.CellMLError, 'cannot be connected to themselves',
            m.add_connection, x, x)
        xsib = a.add_variable('x_sibling', 'liter')
        self.assertRaisesRegex(
            cellml.CellMLError, 'in the same component',
            m.add_connection, x, xsib)

        # Must be sibling or parent/child
        self.assertRaisesRegex(
            cellml.CellMLError, 'siblings or have a parent-child',
            m.add_connection, x, z)

        # Invalid interface
        self.assertRaisesRegex(
            cellml.CellMLError, 'Invalid connection',
            m.add_connection, xsib, y)

        # "in" variable connected to two "out" variables
        m = cellml.Model('m')
        a = m.add_component('a')
        b = m.add_component('b')
        c = m.add_component('c')
        x = a.add_variable('x', 'volt', 'out', 'none')
        y = b.add_variable('y', 'volt', 'out', 'none')
        z = c.add_variable('z', 'volt', 'in', 'none')
        m.add_connection(x, z)
        self.assertRaisesRegex(
            cellml.CellMLError, 'Invalid connection', m.add_connection, x, z)
        self.assertRaisesRegex(
            cellml.CellMLError, 'Invalid connection', m.add_connection, y, z)
        self.assertRaisesRegex(
            cellml.CellMLError, 'Invalid connection', m.add_connection, z, y)

    def test_creation(self):
        # Tests Model creation

        m = cellml.Model('hiya')

        # Test bad name
        self.assertRaisesRegex(
            cellml.CellMLError, 'valid CellML identifier', cellml.Model, '1e2')

        # Test bad version
        self.assertRaisesRegex(
            ValueError, 'supported', cellml.Model, 'm', '2.0')

    def test_name(self):
        # Tests Model.name().

        m = cellml.Model('Stacey')
        self.assertEqual(m.name(), 'Stacey')

    def test_free_variable(self):
        # Tests setting the free variable.

        m = cellml.Model('m')
        c = m.add_component('c')
        x = c.add_variable('x', 'meter')
        y = c.add_variable('y', 'second')
        self.assertFalse(x.is_free())
        self.assertFalse(y.is_free())

        # Test setting
        m.set_free_variable(x)
        self.assertTrue(x.is_free())
        self.assertFalse(y.is_free())

        # Test changing
        m.set_free_variable(y)
        self.assertFalse(x.is_free())
        self.assertTrue(y.is_free())

        # Test unsetting
        m.set_free_variable(None)
        self.assertFalse(x.is_free())
        self.assertFalse(y.is_free())

        # Test wrong model
        z = cellml.Model('mm').add_component('c').add_variable('z', 'liter')
        self.assertRaisesRegex(
            ValueError, 'from this model', m.set_free_variable, z)

        # Variable with in interface
        z1 = c.add_variable('z1', 'second', public_interface='in')
        z2 = c.add_variable('z2', 'second', private_interface='in')
        self.assertRaisesRegex(
            cellml.CellMLError, 'free variable cannot have an "in" interface',
            m.set_free_variable, z1)
        self.assertRaisesRegex(
            cellml.CellMLError, 'free variable cannot have an "in" interface',
            m.set_free_variable, z2)

    def test_sequence_interface(self):
        # Tests the sequence interface on a model

        m = cellml.Model('m')
        a = m.add_component('Stacey')
        x = a.add_variable('x', 'volt')
        y = a.add_variable('y', 'volt')
        z = a.add_variable('z', 'volt')
        self.assertIs(a['x'], x)
        self.assertIs(a['y'], y)
        self.assertIs(a['z'], z)
        self.assertIn('x', a)
        self.assertIn('y', a)
        self.assertIn('z', a)
        self.assertEqual(len(a), 3)

    def test_string_conversion(self):
        # Tests Model.__str__

        m = cellml.Model('mm')
        self.assertEqual(str(m), 'Model[@name="mm"]')

    def test_validation(self):
        # Tests Model.validate(), doesn't check _validate methods

        # More than one variable without a definition
        m = cellml.Model('m')
        c = m.add_component('c')
        x = c.add_variable('x', 'mole')
        y = c.add_variable('y', 'liter')
        with WarningCollector() as w:
            m.validate()
        self.assertIn('No value set for Variable[@name="x"]', w.text())
        self.assertIn('No value set for Variable[@name="y"]', w.text())
        self.assertIn('More than one variable does not have a value', w.text())

        # Free variable has a value, but other value does not
        x.set_initial_value(0.1)
        m.set_free_variable(x)
        with WarningCollector() as w:
            m.validate()
        self.assertIn('No value set for Variable[@name="y"]', w.text())
        self.assertIn('No value is defined for the variable "y"', w.text())

        # Free variable must be known if state is used
        y.set_rhs(myokit.Name(x))
        y.set_initial_value(0.1)
        y.set_is_state(True)
        m.set_free_variable(None)
        self.assertRaisesRegex(
            cellml.CellMLError, 'a free variable must be set', m.validate)

        # Free variable is set, but not present in component containing state
        d = m.add_component('d')
        z = d.add_variable('z', 'second')
        z.set_initial_value(0)
        m.set_free_variable(z)
        self.assertRaisesRegex(
            cellml.CellMLError, 'no local variable connected to the free',
            m.validate)

    def test_version(self):
        # TestsModel.version()

        m = cellml.Model('mm', '1.1')
        self.assertEqual(m.version(), '1.1')


class TestCellML1ModelConversion(unittest.TestCase):
    """
    Tests for converting between Myokit and CellML models.
    """
    def model(self, name=None):
        """ Creates and returns a model with a time variable. """
        m = myokit.Model(name)
        t = m.add_component('env').add_variable('time')
        t.set_rhs(myokit.Number(0, myokit.units.ms))
        t.set_unit(myokit.units.ms)
        t.set_binding('time')
        return m

    def test_m2c_model_names(self):
        # Tests name issues when creating a CellML model

        # Test model name is transferred
        m = self.model('test')
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(cm.name(), 'test')
        m = self.model('test model')
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(cm.name(), 'test_model')

        # Test model is renamed if can't clean name
        m = self.model('123')
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(cm.name(), 'unnamed_myokit_model')

        # Test model is ramed if no name given
        m = self.model()
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(cm.name(), 'unnamed_myokit_model')

        # Test duplicate names are solved
        m = self.model()
        c1 = m.add_component('c1')
        c2 = m.add_component('c2')
        x1 = c1.add_variable('x')
        x2 = c2.add_variable('x')
        y = c2.add_variable('y')
        x1.set_rhs(1)
        x2.set_rhs('3 + c1.x')
        y.set_rhs('2 * x')
        cm = cellml.Model.from_myokit_model(m)
        mm = cm.myokit_model()
        self.assertEqual(mm.get('c2.y').eval(), 8)

    def test_m2c_units(self):
        # Test unit issues when creating a CellML model

        m = self.model()
        c1 = m.add_component('c1')
        c2 = m.add_component('c2')
        x1 = c1.add_variable('x')
        x2 = c2.add_variable('x')
        y = c2.add_variable('y')
        x1.set_rhs(1)
        x2.set_rhs('3 + c1.x')
        y.set_rhs('2 * x')
        cm = cellml.Model.from_myokit_model(m)
        mm = cm.myokit_model()

        # Test numbers and variables without units are dimensionless...
        self.assertEqual(cm['c1']['x'].units().name(), 'dimensionless')
        self.assertEqual(
            cm['c2']['y'].rhs(),
            myokit.Multiply(
                myokit.Number(2, myokit.units.dimensionless),
                myokit.Name(cm['c2']['x']))
        )

        # ...or have a unit inferred from their RHS
        z = c2.add_variable('z')
        z.set_rhs('3 [mole/ms]')
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(
            cm['c2']['z'].units().myokit_unit(),
            myokit.units.mole / myokit.units.ms)

        # ...which should take the time units into account, for state variables
        z.promote(0)
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(
            cm['c2']['z'].units().myokit_unit(), myokit.units.mole)

        # ...even if the time units themselves need to be inferred
        m.get('env.time').set_unit(None)
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(
            cm['c2']['z'].units().myokit_unit(), myokit.units.mole)

        # Units can have fractional exponents
        y.set_rhs('1 [mV] ^ 1.2')
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(
            cm['c2']['y'].units().myokit_unit(), myokit.units.mV**1.2)

        # If a variable doesn't have units, the RHS will be inspected. This can
        # lead to unit errors, which should be ignored
        y.set_rhs('1 [mV] + 3 [A]')
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(
            cm['c2']['y'].units().myokit_unit(), myokit.units.dimensionless)

    def test_m2c_nested_variables(self):
        # Test nested variables are handled, and name conflicts are handled
        # when creating a CellML model.

        m1 = self.model()
        c1 = m1.add_component('c1')
        c2 = m1.add_component('c2')
        a1 = c1.add_variable('a1')
        a2 = a1.add_variable('a2')
        a3 = a2.add_variable('a3')
        a4 = a3.add_variable('a4')
        a3b = c2.add_variable('a3')
        a3b.set_rhs(10)
        a4.set_rhs('1 + c2.a3')
        a3.set_rhs('1 + a4')
        a2.set_rhs('1 + a3')
        a1.set_rhs('1 + a2')
        cm = cellml.Model.from_myokit_model(m1)
        m2 = cm.myokit_model()
        self.assertEqual(m2.get('c1.a1').eval(), 14)

    def test_m2c_derivatives(self):
        # Test derivative support when creating a CellML model

        # Test references to derivatives
        m = self.model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs('log10(1000) + x')
        x.promote(0.1)
        y = c.add_variable('y')
        y.set_rhs('2 * dot(x)')
        cm = cellml.Model.from_myokit_model(m)
        mm = cm.myokit_model()
        self.assertEqual(mm.get('c.y').eval(), 6.2)

    def test_m2c_oxmeta(self):
        # Test that oxmeta data is passed on when creating a CellML model.

        # Test oxmeta data is passed on
        m = self.model()
        c = m.add_component('c')
        x = c.add_variable('x')
        y = x.add_variable('y')
        y.set_rhs(1)
        x.set_rhs('1 + y')
        x.meta['oxmeta'] = 'membrane_voltage'
        y.meta['oxmeta'] = 'fish'
        d = m.add_component('d')
        x2 = d.add_variable('x')
        x2.set_rhs(3)
        cm = cellml.Model.from_myokit_model(m)
        self.assertEqual(len(cm['c']['x'].meta), 1)
        self.assertEqual(len(cm['c']['y'].meta), 1)
        self.assertEqual(len(cm['d']['x'].meta), 0)
        self.assertEqual(cm['c']['x'].meta['oxmeta'], 'membrane_voltage')
        self.assertEqual(cm['c']['y'].meta['oxmeta'], 'fish')

        # Test cmeta id is set if variable has oxmeta annotation
        self.assertEqual(cm['c']['x'].cmeta_id(), 'c_x')
        self.assertEqual(cm['c']['y'].cmeta_id(), 'y')
        self.assertIsNone(cm['d']['x'].cmeta_id())

    def test_evaluating_states(self):
        # Test converting to and from CellML doesn't change the state
        # variable evaluations.

        # Test evaluating states of model
        m = myokit.load_model('example')
        cm = cellml.Model.from_myokit_model(m)
        with WarningCollector() as w:
            cm.validate()
        self.assertFalse(w.has_warnings())

        # Recreate myokit model and test states
        mm = cm.myokit_model()
        mm.validate()
        state_1 = m.initial_values(True)
        state_2 = mm.initial_values(True)
        states_1 = [x.name() for x in m.states()]
        states_2 = [x.name() for x in mm.states()]
        state_2 = [state_2[states_2.index(x)] for x in states_1]
        self.assertEqual(state_1[0], state_2[0])
        self.assertEqual(state_1[1], state_2[1])
        self.assertEqual(state_1[2], state_2[2])
        self.assertEqual(state_1[3], state_2[3])
        self.assertEqual(state_1[4], state_2[4])
        self.assertEqual(state_1[5], state_2[5])
        self.assertEqual(state_1[6], state_2[6])
        self.assertEqual(state_1[7], state_2[7])

    def test_c2m_basic(self):
        # Tests basic features of conversion to a Myokit model

        # Create model
        m = cellml.Model('m')
        documentation = 'This is the documentation.'
        m.meta['documentation'] = documentation
        a = m.add_component('a')
        a.meta['harry'] = 'wilco'
        b = m.add_component('b')
        c = m.add_component('c')
        c.set_parent(b)
        x = a.add_variable('x', 'volt', 'out', 'none')
        xb = b.add_variable('x', 'volt', 'in', 'out')
        xc = c.add_variable('x', 'volt', 'in', 'none')
        xc.meta['blue'] = 'yellow'
        y = b.add_variable('y', 'volt')
        z = c.add_variable('z', 'volt')
        m.add_connection(x, xb)
        m.add_connection(xc, xb)
        z2 = c.add_variable('z2', 'meter')
        z2.set_initial_value(4)
        t = a.add_variable('t', 'second')
        t.meta['yes'] = 'no'
        m.set_free_variable(t)

        x.set_rhs(myokit.Number(1, myokit.units.V))
        x.set_is_state(True)
        x.set_initial_value(0.123)
        y.set_rhs(
            myokit.Plus(myokit.Number(2, myokit.units.V), myokit.Name(xb)))
        z.set_rhs(
            myokit.Plus(myokit.Number(3, myokit.units.V), myokit.Name(xc)))

        # Convert
        mm = m.myokit_model()
        self.assertIsInstance(mm, myokit.Model)
        self.assertEqual(mm.name(), 'm')

        # Check meta data is added
        self.assertIn('author', mm.meta)

        # Check meta data is passed on
        self.assertIn('documentation', mm.meta)
        self.assertEqual(mm.meta['documentation'], documentation)
        self.assertIn('harry', mm.get('a').meta)
        self.assertEqual(mm.get('a').meta['harry'], 'wilco')
        self.assertIn('yes', mm.get('a.t').meta)
        self.assertEqual(mm.get('a.t').meta['yes'], 'no')

        # Check components are present
        ma = mm['a']
        mb = mm['b']
        mc = mm['c']
        self.assertEqual(len(mm), 3)

        # Check variables are present
        mt = ma['t']
        mx = ma['x']
        self.assertEqual(len(ma), 2)
        my = mb['y']
        self.assertEqual(len(mb), 1)
        mz = mc['z']
        mz2 = mc['z2']
        self.assertEqual(len(mc), 2)

        # Check units are set
        self.assertEqual(mt.unit(), myokit.units.second)
        self.assertEqual(mx.unit(), myokit.units.volt)
        self.assertEqual(my.unit(), myokit.units.volt)
        self.assertEqual(mz.unit(), myokit.units.volt)
        self.assertEqual(mz2.unit(), myokit.units.meter)

        # Check RHS equations
        self.assertEqual(mt.rhs(), myokit.Number(0, myokit.units.second))
        self.assertEqual(mx.rhs(), myokit.Number(1, myokit.units.volt))
        self.assertEqual(
            my.rhs(),
            myokit.Plus(myokit.Number(2, myokit.units.volt), myokit.Name(mx)))
        self.assertEqual(
            mz.rhs(),
            myokit.Plus(myokit.Number(3, myokit.units.volt), myokit.Name(mx)))
        self.assertEqual(mz2.rhs(), myokit.Number(4, myokit.units.meter))

        # Check state
        self.assertTrue(mx.is_state())
        self.assertEqual(mx.initial_value(True), 0.123)

        # Check binding
        self.assertEqual(mt.binding(), 'time')

    def test_c2m_pass_through_variables(self):
        # Test support for variables used only to pass a value through a
        # hierarchical CellML structure.

        m = cellml.Model('m')
        a = m.add_component('a')
        b = m.add_component('b')
        c = m.add_component('c')
        d = m.add_component('d')
        c.set_parent(b)
        d.set_parent(c)
        ax = a.add_variable('x', 'volt', 'out', 'none')
        bx = b.add_variable('x', 'volt', 'in', 'out')
        cx = c.add_variable('x', 'volt', 'in', 'out')
        dx = d.add_variable('x', 'volt', 'in', 'none')
        dy = d.add_variable('y', 'volt')
        m.add_connection(ax, bx)
        m.add_connection(cx, bx)
        m.add_connection(dx, cx)
        ax.set_rhs(myokit.Number(1, myokit.units.V))
        dy.set_rhs(myokit.Name(dx))

        # Convert and check
        mm = m.myokit_model()
        self.assertEqual(len(mm), 4)
        self.assertEqual(len(mm['a']), 1)
        self.assertEqual(len(mm['b']), 0)
        self.assertEqual(len(mm['c']), 0)
        self.assertEqual(len(mm['d']), 1)
        self.assertIn('x', mm['a'])
        self.assertIn('y', mm['d'])

    def test_c2m_unit_conversion(self):
        # Test support for unit conversion (and pass-through variables)

        # Create model with encapsulation hierarchy
        cm = cellml.Model('m')
        a = cm.add_component('a')
        b = cm.add_component('b')
        c = cm.add_component('c')
        c.set_parent(b)

        # Add some units
        cm.add_units('millivolt', myokit.units.volt * 1e-3)
        cm.add_units('kilovolt', myokit.units.volt * 1e3)
        cm.add_units('millimole', myokit.units.mole * 1e-3)
        cm.add_units('megamole', myokit.units.mole * 1e6)
        cm.add_units('kilowatt', myokit.units.W * 1e3)

        # x requires conversion, and is used in c: it should get a new variable
        # in c with the correct nuits
        ax = a.add_variable('x', 'volt', 'out', 'none')
        bx = b.add_variable('x', 'millivolt', 'in', 'out')
        cx = c.add_variable('x', 'kilovolt', 'in', 'none')
        ax.set_rhs(myokit.Number(1, myokit.units.volt))

        # y does not require conversion, and is used in c: it should not get a
        # new variable in c
        ay = a.add_variable('y', 'ampere', 'out', 'none')
        by = b.add_variable('y', 'ampere', 'in', 'out')
        cy = c.add_variable('y', 'ampere', 'in', 'none')
        ay.set_rhs(myokit.Number(2, myokit.units.ampere))

        # z requires unit conversion, and is not used in c: it should not get a
        # new variable in c
        az = a.add_variable('z', 'mole', 'out', 'none')
        bz = b.add_variable('z', 'millimole', 'in', 'out')
        cz = c.add_variable('z', 'megamole', 'in', 'none')
        az.set_rhs(myokit.Number(3, myokit.units.mole))

        # p uses variables
        cp = c.add_variable('p', 'kilowatt')
        cp.set_rhs(myokit.Multiply(myokit.Name(cx), myokit.Name(cy)))

        # Connect variables
        cm.add_connection(ax, bx)
        cm.add_connection(bx, cx)
        cm.add_connection(ay, by)
        cm.add_connection(by, cy)
        cm.add_connection(az, bz)
        cm.add_connection(bz, cz)

        # Convert and check
        m = cm.myokit_model()
        a, b, c = m['a'], m['b'], m['c']

        # Check that pass-through variables have disappeared
        self.assertEqual(len(b), 0)

        # Check rest of variables
        self.assertEqual(len(a), 3)
        self.assertEqual(len(c), 2)
        ax, bx, cx = a['x'], a['y'], a['z']
        cx, cp = c['x'], c['p']

        # Check variable RHS's
        self.assertEqual(ax.rhs().code(), '1 [V]')
        self.assertEqual(ay.rhs().code(), '2 [A]')
        self.assertEqual(az.rhs().code(), '3 [mol]')
        self.assertEqual(
            cx.rhs(),
            myokit.Multiply(
                myokit.Name(ax),
                myokit.Number(1e-3, '1 (1000)')
            )
        )
        self.assertEqual(cp.rhs().code(), 'c.x * a.y')

        # Units should now be good
        m.check_units(mode=myokit.UNIT_STRICT)

    def test_c2m_unit_conversion_fail(self):
        # Test warnings are raised if unit conversion fails

        # Create simple model
        cm = cellml.Model('m')
        a = cm.add_component('a')
        b = cm.add_component('b')

        ax = a.add_variable('x', 'volt', 'out', 'none')
        bx = b.add_variable('x', 'ampere', 'in', 'out')
        cm.add_connection(ax, bx)
        ax.set_rhs(myokit.Number(1.23, myokit.units.volt))

        by = b.add_variable('y', 'ampere')
        by.set_rhs(myokit.Name(bx))

        # Convert and check warning was raised
        with WarningCollector() as w:
            m = cm.myokit_model()
        self.assertIn(
            'Unable to determine unit conversion factor for b.x', w.text())

        # Check variable RHS
        ax, bx, by = m.get('a.x'), m.get('b.x'), m.get('b.y')
        self.assertEqual(ax.rhs().code(), '1.23 [V]')
        self.assertEqual(
            bx.rhs(),
            myokit.Multiply(myokit.Name(ax), myokit.Number(1))
        )
        self.assertEqual(by.rhs(), myokit.Name(bx))

        # Units are invalid
        self.assertRaises(myokit.IncompatibleUnitError, m.check_units)


class TestCellML1Variable(unittest.TestCase):
    """ Tests for ``cellml.v1.Variable``. """

    def test_creation(self):
        # Tests variable creation

        m = cellml.Model('mmm')
        c = m.add_component('ccc')
        v = c.add_variable('vee', 'meter')

        # Test setting (and getting) interfaces
        self.assertEqual(v.public_interface(), 'none')
        self.assertEqual(v.private_interface(), 'none')
        w = c.add_variable('doubleyou', 'ampere', 'out', 'in')
        self.assertEqual(w.public_interface(), 'out')
        self.assertEqual(w.private_interface(), 'in')

        # Test bad name
        self.assertRaisesRegex(
            cellml.CellMLError, 'valid CellML identifier',
            c.add_variable, '', 'meter')

        # Test bad units
        self.assertRaisesRegex(
            cellml.CellMLError, 'units attribute',
            c.add_variable, 'soy_sauce', 'meters')

        # Test bad interfaces
        self.assertRaisesRegex(
            cellml.CellMLError, 'Public interface',
            c.add_variable, 'soy', 'meter', 'bin')
        self.assertRaisesRegex(
            cellml.CellMLError, 'Private interface',
            c.add_variable, 'soy', 'meter', 'in', 'bout')
        self.assertRaisesRegex(
            cellml.CellMLError, 'both',
            c.add_variable, 'soy', 'meter', 'in', 'in')

    def test_component_and_model(self):
        # Tests Variable.component() and Variable.model()

        m = cellml.Model('mm')
        c = m.add_component('comp')
        v = c.add_variable('bert', 'meter')
        self.assertIs(v.component(), c)
        self.assertIs(v.model(), m)

    def test_initial_value(self):
        # Tests getting and setting intial values

        v = cellml.Model('m').add_component('c').add_variable('v', 'volt')
        self.assertIsNone(v.initial_value())

        # Test setting and changing
        v.set_initial_value(4)
        self.assertEqual(v.initial_value(), 4)
        v.set_initial_value(-1.2e9)
        self.assertEqual(v.initial_value(), -1.2e9)

        # Test unsetting
        v.set_initial_value(None)
        self.assertIsNone(v.initial_value())

        # Bad value
        self.assertRaisesRegex(
            cellml.CellMLError, 'real number', v.set_initial_value, 'blue')

        # Bad interface
        w = v.component().add_variable('w', 'volt', private_interface='in')
        self.assertRaisesRegex(
            cellml.CellMLError, 'private_interface="in"',
            w.set_initial_value, 1)
        x = v.component().add_variable('x', 'volt', public_interface='in')
        self.assertRaisesRegex(
            cellml.CellMLError, 'public_interface="in"',
            x.set_initial_value, 1)

    def test_is_local_and_source(self):
        # Tests Variable.is_local() and Variable.source() and
        # Variable.value_source()

        m = cellml.Model('m')
        a = m.add_component('a')
        b = m.add_component('b')
        c = m.add_component('c')
        c.set_parent(b)
        ax = a.add_variable('x', 'volt', 'out')
        bx = b.add_variable('x', 'volt', 'in', 'out')
        cx = c.add_variable('x', 'volt', 'in')
        m.add_connection(ax, bx)
        m.add_connection(bx, cx)

        self.assertTrue(ax.is_local())
        self.assertFalse(bx.is_local())
        self.assertFalse(cx.is_local())
        self.assertIs(cx.source(), bx)
        self.assertIs(bx.source(), ax)
        self.assertIs(ax.value_source(), ax)
        self.assertIs(bx.value_source(), ax)
        self.assertIs(cx.value_source(), ax)

    def test_name(self):
        # Tests Variable.name()

        v = cellml.Model('m').add_component('a').add_variable('ernie', 'volt')
        self.assertEqual(v.name(), 'ernie')

    def test_overdefined(self):
        # Tests if overdefined variables are reported

        # State can have initial value and RHS
        m = cellml.Model('m')
        c = m.add_component('c')
        t = c.add_variable('t', 'second')
        v = c.add_variable('v', 'volt')
        v.set_rhs(myokit.Number(2, myokit.units.volt))
        v.set_initial_value(3)
        v.set_is_state(True)
        m.set_free_variable(t)
        m.validate()

        # But non-state can not
        v.set_is_state(False)
        self.assertRaisesRegex(
            cellml.CellMLError, 'both an initial value and a defining',
            m.validate)

    def test_rhs_or_initial_value(self):
        # Tests Variable.rhs_or_initial_value()

        v = cellml.Model('m').add_component('c').add_variable('bert', 'meter')
        self.assertIsNone(v.rhs_or_initial_value())

        # Test initial value is returned
        v.set_initial_value(3)
        i = myokit.Number(3, myokit.units.meter)
        self.assertEqual(v.rhs_or_initial_value(), i)

        # RHS takes precedence over initial value
        r = myokit.Number(18, myokit.units.meter)
        v.set_rhs(r)
        self.assertEqual(v.rhs_or_initial_value(), r)
        v.set_rhs(None)
        self.assertEqual(v.rhs_or_initial_value(), i)

        # State never returns initial value
        v.set_is_state(True)
        self.assertIsNone(v.rhs_or_initial_value())

    def test_set_and_get_rhs(self):
        # Tests Variable.set_rhs() and Variable.rhs()

        v = cellml.Model('m').add_component('c').add_variable('bert', 'meter')
        self.assertIsNone(v.rhs())
        r = myokit.Number(1, myokit.units.meter)
        v.set_rhs(r)
        self.assertEqual(v.rhs(), r)
        v.set_rhs(None)
        self.assertIsNone(v.rhs())

        # Bad interface
        w = v.component().add_variable('w', 'volt', 'in')
        self.assertRaisesRegex(
            cellml.CellMLError, 'public_interface="in"', w.set_rhs, r)
        x = v.component().add_variable('x', 'volt', 'out', 'in')
        self.assertRaisesRegex(
            cellml.CellMLError, 'private_interface="in"', x.set_rhs, r)

        # Invalid references
        z = v.model().add_component('d').add_variable('z', 'volt')
        self.assertRaisesRegex(
            cellml.CellMLError, 'can only reference variables from the same',
            v.set_rhs, myokit.Name(z))

        # Unknown units in RHS
        rhs = myokit.Plus(
            myokit.Number(3),
            myokit.Number(1, myokit.units.volt / myokit.units.mole)
        )
        self.assertRaisesRegex(
            cellml.CellMLError, 'units appearing in a variable\'s RHS',
            v.set_rhs, rhs)

    def test_set_and_is_state(self):
        # Tests Variable.set_is_state() and Variable.is_state()

        m = cellml.Model('m')
        m.add_units('flubber', myokit.units.volt / myokit.units.second)
        c = m.add_component('a')
        v = c.add_variable('v', 'volt')
        self.assertFalse(v.is_state())
        v.set_initial_value(4)
        v.set_rhs(myokit.Number(3, myokit.units.volt / myokit.units.second))
        self.assertFalse(v.is_state())

        # Test setting and unsetting
        v.set_is_state(True)
        self.assertTrue(v.is_state())
        v.set_is_state(False)
        self.assertFalse(v.is_state())

        # Bad interface
        w = v.component().add_variable('w', 'volt', 'in')
        self.assertRaisesRegex(
            cellml.CellMLError, 'an "in" interface', w.set_is_state, True)
        x = v.component().add_variable('x', 'volt', 'out', 'in')
        self.assertRaisesRegex(
            cellml.CellMLError, 'an "in" interface', x.set_is_state, True)

    def test_string_conversion(self):
        # Tests Variable.__str__

        v = cellml.Model('m').add_component('c').add_variable('bert', 'meter')
        self.assertEqual(
            str(v), 'Variable[@name="bert"] in Component[@name="c"]')

    def test_validate(self):
        # Tests Variable validation

        # Unconnected variables
        m = cellml.Model('m')
        a = m.add_component('a')
        b = m.add_component('b')
        t = a.add_variable('t', 'dimensionless', 'out')
        m.set_free_variable(t)
        ax = a.add_variable('x', 'meter', 'in')
        bx = b.add_variable('x', 'meter', 'out')
        bx.set_initial_value(3)
        bt = b.add_variable('toim', 'dimensionless', 'in')
        m.add_connection(t, bt)

        with WarningCollector() as w:
            m.validate()
        self.assertIn('not connected', w.text())

        # States must define two values
        m.add_connection(ax, bx)
        m.validate()
        bx.set_is_state(True)
        self.assertRaisesRegex(
            cellml.CellMLError, 'must have a defining equation', m.validate)
        bx.set_rhs(myokit.Number(1, 'meter'))
        m.validate()
        bx.set_initial_value(None)
        with WarningCollector() as w:
            m.validate()
        self.assertIn('has no initial value', w.text())


class TestCellML1Units(unittest.TestCase):
    """ Tests for ``cellml.Units``. """

    def test_creation(self):
        # Test creation

        mv = cellml.Units('millivolt', myokit.units.mV)
        self.assertRaisesRegex(
            cellml.CellMLError, 'valid CellML identifier',
            cellml.Units, '132', myokit.units.mV)
        self.assertRaisesRegex(
            cellml.CellMLError, 'predefined',
            cellml.Units, 'volt', myokit.units.V)
        self.assertRaisesRegex(
            ValueError, 'myokit.Unit',
            cellml.Units, 'wooster', 'mV')

    def test_find(self):
        # Tests predefined unit lookup

        # Test lookup
        v = cellml.Units.find_units('volt')
        self.assertEqual(v.name(), 'volt')
        self.assertEqual(v.myokit_unit(), myokit.units.volt)

        # Test lookup returns same object on repeated calls
        self.assertIs(v, cellml.Units.find_units('volt'))
        self.assertIs(v, cellml.Units.find_units('volt'))

        # Test bad lookup
        self.assertRaisesRegex(
            cellml.CellMLError, 'Unknown units name',
            cellml.Units.find_units, 'wooster')

        # Test lookup of unsupported units
        self.assertRaises(
            cellml.UnsupportedBaseUnitsError,
            cellml.Units.find_units, 'celsius')

    def test_myokit_unit(self):
        # Tests myokit_unit()

        u = cellml.Units('woppa', myokit.units.pA)
        self.assertEqual(u.myokit_unit(), myokit.units.pA)

    def test_name(self):
        # Tests name()

        u = cellml.Units('woppa', myokit.units.m)
        self.assertEqual(u.name(), 'woppa')

    def test_parse_unit_row(self):
        # Tests Units.parse_unit_row().

        # Create model with model and component units
        m = cellml.Model('m')
        c = m.add_component('c')
        m.add_units('wooster', myokit.units.V)
        c.add_units('wooster', myokit.units.A)

        # Test lookup of predefined, model, and component units
        u = cellml.Units.parse_unit_row('meter')
        self.assertEqual(u, myokit.units.m)
        u = cellml.Units.parse_unit_row('wooster', context=m)
        self.assertEqual(u, myokit.units.V)
        u = cellml.Units.parse_unit_row('wooster', context=c)
        self.assertEqual(u, myokit.units.A)

        # Test bad lookup
        self.assertRaisesRegex(
            cellml.CellMLError, 'Unknown units name',
            cellml.Units.parse_unit_row, 'wooster')
        self.assertRaisesRegex(
            cellml.CellMLError, 'Unknown units name',
            cellml.Units.parse_unit_row, 'muppet', context=c)

        # Unsupported base unit
        u = cellml.Units.parse_unit_row(
            'meter', prefix='milli', exponent=2, multiplier=1.234)
        self.assertRaises(
            cellml.UnsupportedBaseUnitsError,
            cellml.Units.parse_unit_row, 'celsius')
        self.assertRaises(
            cellml.UnsupportedBaseUnitsError,
            cellml.Units.parse_unit_row, 'celsius', context=c)

        # Test prefixes
        u = cellml.Units.parse_unit_row('meter', 'micro')
        self.assertEqual(u, myokit.parse_unit('um'))
        u = cellml.Units.parse_unit_row('meter', -4)
        self.assertEqual(u, myokit.parse_unit('m (1e-4)'))
        u = cellml.Units.parse_unit_row('meter', -4.0)
        self.assertEqual(u, myokit.parse_unit('m (1e-4)'))

        # Test bad prefixes
        self.assertRaisesRegex(
            cellml.CellMLError, 'known prefixes or an integer',
            cellml.Units.parse_unit_row, 'meter', 'forty')
        self.assertRaisesRegex(
            cellml.CellMLError, 'known prefixes or an integer',
            cellml.Units.parse_unit_row, 'meter', 14.8)
        self.assertRaisesRegex(
            cellml.CellMLError, 'too large',
            cellml.Units.parse_unit_row, 'meter', 999)

        # Test exponent
        u = cellml.Units.parse_unit_row('meter', exponent=2)
        self.assertEqual(u, myokit.parse_unit('m^2'))
        u = cellml.Units.parse_unit_row('meter', exponent=-1.0)
        self.assertEqual(u, myokit.parse_unit('m^-1'))

        # Test bad exponent
        self.assertRaisesRegex(
            cellml.CellMLError, 'must be a real number',
            cellml.Units.parse_unit_row, 'meter', exponent='bert')

        # Test multiplier
        u = cellml.Units.parse_unit_row('meter', multiplier=1.234)
        self.assertEqual(u, myokit.parse_unit('m (1.234)'))

        # Bad multiplier
        self.assertRaisesRegex(
            cellml.CellMLError, 'must be a real number',
            cellml.Units.parse_unit_row, 'meter', multiplier='hiya')

        # All combined
        u = cellml.Units.parse_unit_row(
            'meter', prefix='milli', exponent=2, multiplier=1.234)
        self.assertEqual(u, myokit.parse_unit('mm^2 (1.234)'))

    def test_predefined(self):
        # Tests the predefined units exist and map to the correct myokit units

        self.assertEqual(cellml.Units.find_units(
            'ampere').myokit_unit(), myokit.units.A)
        self.assertEqual(cellml.Units.find_units(
            'becquerel').myokit_unit(), myokit.units.Bq)
        self.assertEqual(cellml.Units.find_units(
            'candela').myokit_unit(), myokit.units.cd)
        self.assertEqual(cellml.Units.find_units(
            'coulomb').myokit_unit(), myokit.units.C)
        self.assertEqual(cellml.Units.find_units(
            'dimensionless').myokit_unit(), myokit.units.dimensionless)
        self.assertEqual(cellml.Units.find_units(
            'farad').myokit_unit(), myokit.units.F)
        self.assertEqual(cellml.Units.find_units(
            'gram').myokit_unit(), myokit.units.g)
        self.assertEqual(cellml.Units.find_units(
            'gray').myokit_unit(), myokit.units.Gy)
        self.assertEqual(cellml.Units.find_units(
            'henry').myokit_unit(), myokit.units.H)
        self.assertEqual(cellml.Units.find_units(
            'hertz').myokit_unit(), myokit.units.Hz)
        self.assertEqual(cellml.Units.find_units(
            'joule').myokit_unit(), myokit.units.J)
        self.assertEqual(cellml.Units.find_units(
            'katal').myokit_unit(), myokit.units.kat)
        self.assertEqual(cellml.Units.find_units(
            'kelvin').myokit_unit(), myokit.units.K)
        self.assertEqual(cellml.Units.find_units(
            'kilogram').myokit_unit(), myokit.units.kg)
        self.assertEqual(cellml.Units.find_units(
            'liter').myokit_unit(), myokit.units.L)
        self.assertEqual(cellml.Units.find_units(
            'litre').myokit_unit(), myokit.units.L)
        self.assertEqual(cellml.Units.find_units(
            'lumen').myokit_unit(), myokit.units.lm)
        self.assertEqual(cellml.Units.find_units(
            'lux').myokit_unit(), myokit.units.lux)
        self.assertEqual(cellml.Units.find_units(
            'meter').myokit_unit(), myokit.units.m)
        self.assertEqual(cellml.Units.find_units(
            'metre').myokit_unit(), myokit.units.m)
        self.assertEqual(cellml.Units.find_units(
            'mole').myokit_unit(), myokit.units.mol)
        self.assertEqual(cellml.Units.find_units(
            'newton').myokit_unit(), myokit.units.N)
        self.assertEqual(cellml.Units.find_units(
            'ohm').myokit_unit(), myokit.units.R)
        self.assertEqual(cellml.Units.find_units(
            'pascal').myokit_unit(), myokit.units.Pa)
        self.assertEqual(cellml.Units.find_units(
            'radian').myokit_unit(), myokit.units.rad)
        self.assertEqual(cellml.Units.find_units(
            'second').myokit_unit(), myokit.units.s)
        self.assertEqual(cellml.Units.find_units(
            'siemens').myokit_unit(), myokit.units.S)
        self.assertEqual(cellml.Units.find_units(
            'sievert').myokit_unit(), myokit.units.Sv)
        self.assertEqual(cellml.Units.find_units(
            'steradian').myokit_unit(), myokit.units.sr)
        self.assertEqual(cellml.Units.find_units(
            'tesla').myokit_unit(), myokit.units.T)
        self.assertEqual(cellml.Units.find_units(
            'volt').myokit_unit(), myokit.units.V)
        self.assertEqual(cellml.Units.find_units(
            'watt').myokit_unit(), myokit.units.W)
        self.assertEqual(cellml.Units.find_units(
            'weber').myokit_unit(), myokit.units.Wb)

    def test_prefixes(self):
        # Tests if all units prefixes are parsed correctly.

        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'yotta'),
            myokit.parse_unit('m (1e24)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'zetta'),
            myokit.parse_unit('m (1e21)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'exa'),
            myokit.parse_unit('m (1e18)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'peta'),
            myokit.parse_unit('m (1e15)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'tera'),
            myokit.parse_unit('m (1e12)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'giga'),
            myokit.parse_unit('m (1e9)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'mega'),
            myokit.parse_unit('m (1e6)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'kilo'),
            myokit.parse_unit('m (1e3)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'hecto'),
            myokit.parse_unit('m (1e2)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'deka'),
            myokit.parse_unit('m (1e1)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'deca'),
            myokit.parse_unit('m (1e1)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'deci'),
            myokit.parse_unit('m (1e-1)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'centi'),
            myokit.parse_unit('m (1e-2)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'milli'),
            myokit.parse_unit('m (1e-3)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'micro'),
            myokit.parse_unit('m (1e-6)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'nano'),
            myokit.parse_unit('m (1e-9)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'pico'),
            myokit.parse_unit('m (1e-12)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'femto'),
            myokit.parse_unit('m (1e-15)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'atto'),
            myokit.parse_unit('m (1e-18)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'zepto'),
            myokit.parse_unit('m (1e-21)'))
        self.assertEqual(
            cellml.Units.parse_unit_row('meter', 'yocto'),
            myokit.parse_unit('m (1e-24)'))

    def test_si_unit_names(self):
        # Tests Units.si_unit_names()

        names = [x for x in cellml.Units.si_unit_names()]
        self.assertTrue(len(names) > 10)
        for name in names:
            self.assertTrue(isinstance(name, str))
            if name == 'celsius':
                continue
            self.assertTrue(
                isinstance(cellml.Units.find_units(name), cellml.Units))

    def test_string_conversion(self):
        # Test __str__

        u = cellml.Units('Numpty', myokit.units.ampere)
        self.assertEqual(str(u), 'Units[@name="Numpty"]')


class TestCellML1Methods(unittest.TestCase):
    """ Tests for public CellML API methods. """

    def test_clean_identifier(self):
        # Tests clean_identifier().

        # Valid identifiers get passed through
        name = 'hello'
        self.assertEqual(cellml.clean_identifier(name), name)
        name = 'h_ello'
        self.assertEqual(cellml.clean_identifier(name), name)
        name = 'x123'
        self.assertEqual(cellml.clean_identifier(name), name)
        name = 'x_123'
        self.assertEqual(cellml.clean_identifier(name), name)

        # Some identifiers can be fixed
        self.assertEqual(
            cellml.clean_identifier('this is a model'),
            'this_is_a_model')
        self.assertEqual(
            cellml.clean_identifier('This-is-a-model'),
            'This_is_a_model')

        # Some identifiers can't be fixed
        self.assertRaises(ValueError, cellml.clean_identifier, '18 + 25')

    def test_create_unit_name(self):
        # Tests create_unit_name().

        # Easy ones
        u = myokit.units.kg
        self.assertEqual(cellml.create_unit_name(u), 'kg')
        u = myokit.units.mV
        self.assertEqual(cellml.create_unit_name(u), 'mV')

        # Easy unit with funny multiplier
        u = myokit.units.m * 1.234
        self.assertEqual(cellml.create_unit_name(u), 'm_times_1_dot_234')

        # Dimensionless unit with multiplier
        u = myokit.units.dimensionless * 2
        self.assertEqual(cellml.create_unit_name(u), 'dimensionless_times_2')

        # Dimensionless unit with multiplier in e-notation
        u = myokit.units.dimensionless * 1000
        self.assertEqual(cellml.create_unit_name(u), 'dimensionless_times_1e3')
        u = myokit.units.dimensionless / 1000
        self.assertEqual(
            cellml.create_unit_name(u), 'dimensionless_times_1e_minus_3')

        # Per meter
        u = 1 / myokit.units.meter
        self.assertEqual(cellml.create_unit_name(u), 'per_m')

        # Unit known to myokit (per second
        u = 1 / myokit.units.second
        self.assertEqual(cellml.create_unit_name(u), 'S_per_F')

        # Square meters
        u = myokit.units.meter ** 2
        self.assertEqual(cellml.create_unit_name(u), 'm2')

    def test_valid_identifier(self):
        # Tests is_valid_identifier().

        self.assertTrue(cellml.is_valid_identifier('hello'))
        self.assertTrue(cellml.is_valid_identifier('h_e_l_l_o'))
        self.assertTrue(cellml.is_valid_identifier('X123'))
        self.assertTrue(cellml.is_valid_identifier('ZAa123_lo_2'))
        self.assertTrue(cellml.is_valid_identifier('a'))
        self.assertTrue(cellml.is_valid_identifier('_a'))

        self.assertFalse(cellml.is_valid_identifier('_'))
        self.assertFalse(cellml.is_valid_identifier('123'))
        self.assertFalse(cellml.is_valid_identifier('1e3'))
        self.assertFalse(cellml.is_valid_identifier('3'))
        self.assertFalse(cellml.is_valid_identifier(''))


if __name__ == '__main__':
    import warnings
    warnings.simplefilter('always')
    unittest.main()



#!/usr/bin/env python3
#
# Tests the Variable class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import numpy as np

import myokit

from myokit.tests import WarningCollector


class VariableTest(unittest.TestCase):
    """
    Tests parts of :class:`myokit.Variable`.
    """

    def test_clamp(self):
        # Tests clamping a variable to a fixed value
        m = myokit.parse_model("""
            [[model]]
            z.g = 2
            z.h = 6.123
            z.p = 1.5

            [z]
            t = 0 [s]
                in [s]
                bind time
            a = 3 [ms]
                in [ms]
            b = 4 [mV]
            b2 = 3
                in [A]
            c = 4 [A] * 3 [A] / 2 [A] + 5 [A]
                in [A]
            d = e + f + t + 8 [m*s] / g
                e = 1 [s]
                    in [s]
                f = 2 [s]
                    in [s]
                in [s]
            dot(g) = 3 [m/s]
                in [m]
            dot(h) = h / 3 [s] + i + 2 [m] / d
                i = 4 [m/s] * j * m * n + h / 1 [s]
                    in [m/s]
                    j = 3 + 2 * k + n
                        k = 4
                    m = j - 1
                    n = 2
                in [m]
            dot(p) = 1
                in [mV]
            """)
        m.validate()
        m.check_units(myokit.UNIT_TOLERANT)
        m_org = m.clone()

        # Clamp literal constants
        m.get('z.a').clamp(5)
        self.assertEqual(m.get('z.a').rhs(), myokit.Number(5, myokit.units.ms))
        m.get('z.b').clamp()
        self.assertEqual(m.get('z.b').rhs(), myokit.Number(4))
        m.get('z.b2').clamp(1)
        self.assertEqual(m.get('z.b2').rhs(), myokit.Number(1, myokit.units.A))
        m.get('z.c').clamp()
        self.assertEqual(m.get('z.c').rhs(), myokit.Number(11, myokit.units.A))

        # Clamp variable with child and sibling dependencies
        d = m.get('z.d')
        self.assertEqual(len(d), 2)
        m.get('z.d.e')
        self.assertFalse(d.is_constant())
        d.clamp()
        self.assertEqual(d.rhs(), myokit.Number(7, myokit.units.s))
        self.assertTrue(d.is_constant())
        self.assertEqual(len(d), 0)
        self.assertRaises(KeyError, m.get, 'z.d.e')
        self.assertRaises(KeyError, m.get, 'z.d.f')
        del d

        # Clamp simple state
        g = m.get('z.g')
        self.assertTrue(g.is_state())
        self.assertNotEqual(g.initial_value(True), 10)
        self.assertEqual(g.unit(), myokit.units.m)
        self.assertEqual(g.rhs().unit(), myokit.units.m / myokit.units.s)
        self.assertEqual(m.count_states(), 3)
        g.clamp(10)
        self.assertTrue(g.is_literal())
        self.assertFalse(g.is_state())
        self.assertEqual(g.rhs(), myokit.Number(10, myokit.units.m))
        self.assertEqual(m.count_states(), 2)
        del g

        # Clamp simple state with missing units
        p = m.get('z.p')
        self.assertTrue(p.is_state())
        self.assertTrue(p.rhs().eval() != 1.5)
        self.assertEqual(m.count_states(), 2)
        p.clamp()
        self.assertTrue(p.is_literal())
        self.assertFalse(p.is_state())
        self.assertEqual(p.rhs(), myokit.Number(1.5, myokit.units.mV))
        self.assertEqual(m.count_states(), 1)
        del p

        # Clamp state with child and sibling dependencies
        h = m.get('z.h')
        self.assertEqual(len(h), 1)
        self.assertEqual(len(list(h.variables(deep=True))), 5)
        m.get('z.h.i.j.k')
        self.assertFalse(h.is_constant())
        self.assertEqual(m.count_states(), 1)
        h.clamp()
        self.assertEqual(h.rhs(), myokit.Number(6.123, myokit.units.m))
        self.assertTrue(h.is_constant())
        self.assertEqual(len(h), 0)
        self.assertRaises(KeyError, m.get, 'z.h.i.j.k')
        self.assertRaises(KeyError, m.get, 'z.h.i')
        self.assertRaises(KeyError, m.get, 'z.h.i.m')
        self.assertEqual(m.count_states(), 0)
        del h

        # Clamp nested variable
        m = m_org.clone()
        h = m.get('z.h')
        x = h.eval()
        self.assertEqual(len(list(h.variables(deep=True))), 5)
        self.assertEqual(len(h), 1)
        h.get('i.j.k')
        self.assertFalse(h.is_constant())
        i = h.get('i')
        self.assertEqual(len(list(i.variables(deep=True))), 4)
        self.assertFalse(i.is_literal())
        y = i.eval()
        i.clamp()
        self.assertTrue(i.is_literal())
        self.assertEqual(i.eval(), y)
        self.assertEqual(h.eval(), x)
        self.assertEqual(len(list(i.variables(deep=True))), 0)
        self.assertEqual(len(list(h.variables(deep=True))), 1)
        del h, i

    def test_convert_unit(self):
        # Test changing variable units

        m = myokit.parse_model("""
            [[model]]
            membrane.V = -83
            membrane.w = 10

            [env]
            t = 0 [ms] bind time
                in [ms]

            [membrane]
            dot(V) = - (ikr.i + ina.i * 1 [cm^2/uF])
                in [mV]
            dotv = 5 [mV/ms] + dot(V)
                in [mV/ms]
            dot(w) = 0.1 [mg/ms]
                in [mg]

            [cell]
            Cm = 0.123 [uF]
                in [uF]

            [ina]
            i = 12 [uA/cm^2]
                in [uA/cm^2]

            [ikr]
            use membrane.V
            E = -81 [mV]
                in [mV]
            g = 23 [mS/uF]
                in [mS/uF]
            i = g * (V - E)
                in [uA/uF]
        """)
        m.check_units(myokit.UNIT_STRICT)

        # Convert to same
        code = m.code()
        v = m.get('membrane.V')
        v.convert_unit('mV')
        self.assertEqual(code, m.code())

        # Convert state (note: initial value is treated as PrefixMinus expr!)
        vdot = v.rhs().eval()
        self.assertNotEqual(v.unit(), myokit.units.V)
        v.convert_unit('V')
        self.assertEqual(v.unit(), myokit.units.V)

        # Check dot(v) is the same, and all units are compatible
        vdot /= 1000
        self.assertEqual(vdot, v.rhs().eval())
        m.check_units(myokit.UNIT_STRICT)

        # Check units of the initial value have been converted
        self.assertEqual(v.initial_value(True), -0.083)
        # Not sure if this one is desirable, but -83 is parsed as the
        # expression -(83)
        self.assertEqual(
            v.initial_value(),
            myokit.Multiply(
                myokit.PrefixMinus(myokit.Number(83)),
                myokit.Number(0.001, myokit.units.V / myokit.units.mV)))

        # Convert state with number initial value
        w = m.get('membrane.w')
        wdot = w.rhs().eval()
        w.convert_unit('ug')
        self.assertEqual(w.unit(), myokit.units.g / 1e6)
        self.assertEqual(wdot * 1000, w.rhs().eval())
        m.check_units(myokit.UNIT_STRICT)
        self.assertEqual(w.initial_value(True), 10000)
        self.assertEqual(w.initial_value(), myokit.Number(10000))
        #self.assertEqual(
        #    w.initial_value(), myokit.Number(10000, myokit.units.g / 1e6))

        # Convert non-state
        i = m.get('ina.i')
        self.assertNotEqual(i.unit(), myokit.parse_unit('uA/uF'))
        i.convert_unit('uA/uF', ['1 [uF/cm^2]'])
        self.assertEqual(i.unit(), myokit.parse_unit('uA/uF'))

        # Check dot(v) is the same, and all units are compatible
        self.assertEqual(vdot, v.rhs().eval())
        m.check_units(myokit.UNIT_STRICT)

        # Convert time variable
        t = m.time()
        self.assertNotEqual(t.unit(), myokit.units.s)
        t.convert_unit('s')
        self.assertEqual(t.unit(), myokit.units.s)

        # Check dot(v) is the same, and all units are compatible
        vdot *= 1000
        self.assertEqual(vdot, v.rhs().eval())
        m.check_units(myokit.UNIT_STRICT)

    def test_index(self):
        # Tests the index() method

        m = myokit.Model()
        c = m.add_component('c')
        a = c.add_variable('a')
        a.set_rhs(0)
        a.promote(0)
        b = c.add_variable('b')
        b.set_rhs(0)
        b.promote(0)

        self.assertEqual(a.index(), 0)
        self.assertEqual(b.index(), 1)
        m.reorder_state((b, a))
        self.assertEqual(a.index(), 1)
        self.assertEqual(b.index(), 0)

        # Test deprecated alias
        with WarningCollector() as w:
            self.assertEqual(a.indice(), 1)
        self.assertIn('deprecated', w.text())

    def test_initial_value(self):
        # Tests :meth:`Variable.initial_value`.

        m = myokit.parse_model('''
            [[model]]
            c.x = 2

            [c]
            t = 0 bind time
            dot(x) = 0
            p = 3
            q = 4
        ''')

        # Test expression version
        x, p = m.get('c.x'), m.get('c.p')
        p = m.get('c.p')
        self.assertEqual(x.initial_value(), myokit.Number(2))
        x.set_initial_value(3)
        self.assertEqual(x.initial_value(), myokit.Number(3))
        x.set_initial_value('3 [kg]')
        self.assertEqual(x.initial_value(), myokit.Number(3, myokit.units.kg))
        x.set_initial_value('sqrt(c.p) / log(c.p)')
        self.assertEqual(x.initial_value(), myokit.Divide(
            myokit.Sqrt(p.lhs()), myokit.Log(p.lhs())))

        # Test float version
        p.set_rhs(9)
        x.set_initial_value('1 + sqrt(c.p)')
        self.assertEqual(x.initial_value(True), 4)

        # Test deprecated alias
        with WarningCollector() as w:
            self.assertEqual(x.state_value(), x.initial_value(True))
        self.assertIn('deprecated', w.text())

        # Test cycles are detected
        p.set_rhs('1 - 2 * q')
        m.get('c.q').set_rhs('7 * sqrt(p)')
        x.initial_value()  # No evaluation, so no error
        self.assertRaises(
            myokit.CyclicalDependencyError, x.initial_value, as_float=True)

    def test_is_referenced(self):
        # Test :meth:`Variable.is_referenced().

        m = myokit.Model()
        c = m.add_component('c')
        v = c.add_variable('v')
        v.set_rhs(3)
        w = c.add_variable('w')
        w.set_rhs(4)

        self.assertFalse(v.is_referenced())
        self.assertFalse(w.is_referenced())
        v.set_rhs('3 * w')
        self.assertFalse(v.is_referenced())
        self.assertTrue(w.is_referenced())
        w.set_rhs('2 * v')
        self.assertTrue(v.is_referenced())
        self.assertTrue(w.is_referenced())

        z = c.add_variable('z')
        z.set_rhs('3 * v')
        self.assertFalse(z.is_referenced())
        self.assertTrue(v.is_referenced())
        w.set_rhs(1)
        self.assertTrue(v.is_referenced())
        z.set_rhs(2)
        self.assertFalse(v.is_referenced())

    def test_labelling(self):
        # Test variable labelling.

        m = myokit.Model()
        c = m.add_component('c')
        v = c.add_variable('v')
        v.set_rhs(3)

        self.assertFalse(v.is_labelled())
        self.assertIsNone(v.label())
        v.set_label('membrane_potential')
        self.assertTrue(v.is_labelled())
        self.assertEqual(v.label(), 'membrane_potential')
        v.set_label(None)
        self.assertFalse(v.is_labelled())
        self.assertIsNone(v.label())

    def test_promote_demote(self):
        # Test variable promotion and demotion.

        m = myokit.Model()
        c = m.add_component('c')
        p = c.add_variable('p')
        p.set_rhs(1)
        v = c.add_variable('v')
        v.set_rhs(3)

        self.assertTrue(v.is_literal())
        self.assertTrue(v.is_constant())
        self.assertFalse(v.is_intermediary())
        self.assertFalse(v.is_state())
        self.assertEqual(v.lhs(), myokit.Name(v))
        self.assertRaises(Exception, v.demote)
        self.assertRaises(Exception, v.index)
        self.assertRaises(Exception, v.initial_value)

        v.promote(3)
        self.assertFalse(v.is_literal())
        self.assertFalse(v.is_constant())
        self.assertFalse(v.is_intermediary())
        self.assertTrue(v.is_state())
        self.assertEqual(v.lhs(), myokit.Derivative(myokit.Name(v)))
        self.assertEqual(v.index(), 0)
        self.assertEqual(v.initial_value(), myokit.Number(3))

        v.demote()
        self.assertTrue(v.is_literal())
        self.assertTrue(v.is_constant())
        self.assertFalse(v.is_intermediary())
        self.assertFalse(v.is_state())
        self.assertEqual(v.lhs(), myokit.Name(v))
        self.assertRaises(Exception, v.demote)
        self.assertRaises(Exception, v.index)
        self.assertRaises(Exception, v.initial_value)

        # Test errors
        v.promote(3)
        self.assertRaisesRegex(Exception, 'already', v.promote, 4)
        v.demote()
        v.set_binding('time')
        self.assertRaisesRegex(Exception, 'cannot be bound', v.promote, 4)
        w = v.add_variable('w')
        self.assertRaisesRegex(
            Exception, 'only be added to Components', w.promote, 4)

        # Test that promote accepts expressions and strings
        v.set_binding(None)
        v.remove_variable(w)
        v.promote(myokit.Number(3, myokit.units.kg))
        self.assertEqual(v.initial_value(), myokit.Number(3, myokit.units.kg))
        v.demote()
        v.promote('1 + c.p')
        self.assertEqual(v.initial_value().code(), '1 + c.p')
        v.demote()

        # Test that promote only works with global syntax references
        self.assertRaisesRegex(myokit.ParseError, 'Unresolved', v.promote, 'p')

        # Test we can't demote a variable with references to its derivative
        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs(3)
        x.promote()
        y = c.add_variable('y')
        y.set_rhs('1 + dot(x)')
        self.assertRaisesRegex(
            Exception, 'references to its derivative', x.demote)
        y.set_rhs('1 + x')
        x.demote()

        # Test deprecated keyword argument
        with WarningCollector() as w:
            x.promote(state_value=3)
        self.assertIn('deprecated', w.text())
        self.assertEqual(x.initial_value(), myokit.Number(3))
        x.demote()
        self.assertRaisesRegex(
            Exception, 'at the same time as', x.promote, 3, state_value=3)

    def test_pyfunc(self):
        # Test :meth:`Variable.pyfunc().

        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs(3)
        y = c.add_variable('y')
        y.set_rhs(4)
        z = c.add_variable('z')
        z.set_rhs('3 * x + y')

        # Invalid model: pyfunc fails
        self.assertRaises(myokit.IntegrityError, z.pyfunc)
        t = c.add_variable('time')
        t.set_rhs(0)
        t.set_binding('time')

        # No states --> No arguments
        f = z.pyfunc(use_numpy=False)
        self.assertEqual(f(), 13)
        f = z.pyfunc(use_numpy=True)
        self.assertEqual(f(), 13)
        f, args = z.pyfunc(use_numpy=False, arguments=True)
        self.assertEqual(args, [])
        f, args = z.pyfunc(use_numpy=True, arguments=True)
        self.assertEqual(args, [])

        # One state
        y.promote(3)
        f = z.pyfunc(use_numpy=False)
        self.assertEqual(f(1), 10)
        f = z.pyfunc(use_numpy=True)
        self.assertEqual(f(1), 10)
        self.assertTrue(
            np.all(f(np.array([1, 2, 4])) == np.array([10, 11, 13])))
        f, args = z.pyfunc(use_numpy=False, arguments=True)
        self.assertEqual(args, (myokit.Name(y), ))
        f, args = z.pyfunc(use_numpy=True, arguments=True)
        self.assertEqual(args, (myokit.Name(y), ))

        # Two states (alphabetically ordered)
        x.promote(2)
        f = z.pyfunc(use_numpy=False)
        self.assertEqual(f(1, 2), 5)
        f = z.pyfunc(use_numpy=True)
        self.assertEqual(f(1, 2), 5)
        self.assertTrue(
            np.all(f(
                np.array([1, 2, 4]), np.array([3, 2, 1])
            ) == np.array([6, 8, 13]))
        )
        f, args = z.pyfunc(use_numpy=False, arguments=True)
        self.assertEqual(args, (myokit.Name(x), myokit.Name(y)))
        f, args = z.pyfunc(use_numpy=True, arguments=True)
        self.assertEqual(args, (myokit.Name(x), myokit.Name(y)))

    def test_refs_by_and_to(self):
        # Test :meth:`Variable.is_referenced().

        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs(3)
        y = c.add_variable('y')
        y.set_rhs(4)
        z = c.add_variable('z')
        z.set_rhs(0)

        self.assertEqual(list(x.refs_to()), [])
        self.assertEqual(list(y.refs_to()), [])
        self.assertEqual(list(z.refs_to()), [])
        self.assertEqual(list(x.refs_by()), [])
        self.assertEqual(list(y.refs_by()), [])
        self.assertEqual(list(z.refs_by()), [])

        z.set_rhs('3 * x')
        self.assertEqual(list(x.refs_to()), [])
        self.assertEqual(list(y.refs_to()), [])
        self.assertEqual(list(z.refs_to()), [x])
        self.assertEqual(list(x.refs_by()), [z])
        self.assertEqual(list(y.refs_by()), [])
        self.assertEqual(list(z.refs_by()), [])

        y.set_rhs('2 + x')
        self.assertEqual(list(x.refs_to()), [])
        self.assertEqual(list(y.refs_to()), [x])
        self.assertEqual(list(z.refs_to()), [x])
        self.assertEqual(set(x.refs_by()), set([y, z]))
        self.assertEqual(list(y.refs_by()), [])
        self.assertEqual(list(z.refs_by()), [])

        # Remove by removing RHS
        y.set_rhs(None)
        self.assertEqual(list(x.refs_to()), [])
        self.assertEqual(list(y.refs_to()), [])
        self.assertEqual(list(z.refs_to()), [x])
        self.assertEqual(set(x.refs_by()), set([z]))
        self.assertEqual(list(y.refs_by()), [])
        self.assertEqual(list(z.refs_by()), [])
        y.set_rhs('2 + x')

        # Remove by changing RHS
        z.set_rhs(2)
        self.assertEqual(list(x.refs_to()), [])
        self.assertEqual(list(y.refs_to()), [x])
        self.assertEqual(list(z.refs_to()), [])
        self.assertEqual(list(x.refs_by()), [y])
        self.assertEqual(list(y.refs_by()), [])
        self.assertEqual(list(z.refs_by()), [])

        # State refs
        self.assertRaises(Exception, x.refs_by, True)
        self.assertRaises(Exception, y.refs_by, True)
        self.assertRaises(Exception, z.refs_by, True)
        self.assertEqual(list(x.refs_to(True)), [])
        self.assertEqual(list(y.refs_to(True)), [])
        self.assertEqual(list(z.refs_to(True)), [])

        # After promoting x, its refs should become srefs
        x.promote(3)
        self.assertEqual(list(x.refs_to(False)), [])
        self.assertEqual(list(y.refs_to(False)), [])
        self.assertEqual(list(z.refs_to(False)), [])
        self.assertEqual(list(x.refs_by(False)), [])
        self.assertEqual(list(y.refs_by(False)), [])
        self.assertEqual(list(z.refs_by(False)), [])
        self.assertEqual(list(x.refs_to(True)), [])
        self.assertEqual(list(y.refs_to(True)), [x])
        self.assertEqual(list(z.refs_to(True)), [])
        self.assertEqual(list(x.refs_by(True)), [y])
        self.assertRaises(Exception, y.refs_by, True)
        self.assertRaises(Exception, z.refs_by, True)

        # Add another reference to x, should now appear in state refs
        z.set_rhs('3 + x')
        self.assertEqual(list(x.refs_to(False)), [])
        self.assertEqual(list(y.refs_to(False)), [])
        self.assertEqual(list(z.refs_to(False)), [])
        self.assertEqual(list(x.refs_by(False)), [])
        self.assertEqual(list(y.refs_by(False)), [])
        self.assertEqual(list(z.refs_by(False)), [])
        self.assertEqual(list(x.refs_to(True)), [])
        self.assertEqual(list(y.refs_to(True)), [x])
        self.assertEqual(list(z.refs_to(True)), [x])
        self.assertEqual(set(x.refs_by(True)), set([y, z]))
        self.assertRaises(Exception, y.refs_by, True)
        self.assertRaises(Exception, z.refs_by, True)
        self.assertEqual(list(y._srefs_by), [])
        self.assertEqual(list(z._srefs_by), [])

        # Demote x: Now its state refs should become ordinary refs again
        x.demote()
        self.assertEqual(list(x.refs_to(False)), [])
        self.assertEqual(list(y.refs_to(False)), [x])
        self.assertEqual(list(z.refs_to(False)), [x])
        self.assertEqual(set(x.refs_by(False)), set([y, z]))
        self.assertEqual(list(y.refs_by(False)), [])
        self.assertEqual(list(z.refs_by(False)), [])
        self.assertEqual(list(x.refs_to(True)), [])
        self.assertEqual(list(y.refs_to(True)), [])
        self.assertEqual(list(z.refs_to(True)), [])
        self.assertRaises(Exception, x.refs_by, True)
        self.assertRaises(Exception, y.refs_by, True)
        self.assertRaises(Exception, z.refs_by, True)
        self.assertEqual(list(x._srefs_by), [])
        self.assertEqual(list(y._srefs_by), [])
        self.assertEqual(list(z._srefs_by), [])
        x.validate()

        #
        # Another test, this time promoting first, then demoting later
        #
        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.promote(0)
        x.set_rhs(3)
        y = c.add_variable('y')
        y.set_rhs('3 + x')

        self.assertEqual(list(x.refs_to(False)), [])
        self.assertEqual(list(y.refs_to(False)), [])
        self.assertEqual(list(x.refs_by(False)), [])
        self.assertEqual(list(y.refs_by(False)), [])
        self.assertEqual(list(x.refs_to(True)), [])
        self.assertEqual(list(y.refs_to(True)), [x])
        self.assertEqual(list(x.refs_by(True)), [y])
        self.assertRaises(Exception, y.refs_by, True)

        # Demote x: Now its state refs should become ordinary refs again
        x.demote()
        self.assertEqual(list(x.refs_to(False)), [])
        self.assertEqual(list(y.refs_to(False)), [x])
        self.assertEqual(list(x.refs_by(False)), [y])
        self.assertEqual(list(y.refs_by(False)), [])
        self.assertEqual(list(x.refs_to(True)), [])
        self.assertEqual(list(y.refs_to(True)), [])
        self.assertRaises(Exception, x.refs_by, True)
        self.assertRaises(Exception, y.refs_by, True)
        x.validate()

        #
        # Another test, this time including self reference by a state
        #
        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        x = c.add_variable('x')
        x.promote(0)
        x.set_rhs('3 - x')
        y = c.add_variable('y')
        y.set_rhs('3 + x')

        self.assertEqual(list(x.refs_to(False)), [])
        self.assertEqual(list(y.refs_to(False)), [])
        self.assertEqual(list(x.refs_by(False)), [])
        self.assertEqual(list(y.refs_by(False)), [])
        self.assertEqual(list(x.refs_to(True)), [x])
        self.assertEqual(list(y.refs_to(True)), [x])
        self.assertEqual(set(x.refs_by(True)), set([x, y]))
        self.assertRaises(Exception, y.refs_by, True)

        # Now demoting causes self-reference
        m.validate()
        x.demote()
        x.validate()
        self.assertRaises(myokit.CyclicalDependencyError, m.validate)

    def test_remove_child_variables(self):
        # Tests removing all child variables
        m = myokit.parse_model("""
            [[model]]

            [z]
            t = 0 [s]
                in [s]
                bind time
            a = 4 [A] * 3 [A] / 2 [A] + 5 [A]
                in [A]
            b = c + d + t + 8 [s]
                c = 1 [s]
                    in [s]
                d = 2 [s]
                    in [s]
                in [s]
            e = f
                f = g
                    g = h
                    h = i + j
                        i = 1 [m]
                        j = 2 [m]
                in [m]
            """)
        m.validate()
        m.check_units(myokit.UNIT_TOLERANT)
        m_org = m.clone()

        # Remove children on var without children
        m.get('z.a').remove_child_variables()
        self.assertTrue(m.is_similar(m_org, True))

        # Remove children on var with used children
        self.assertRaisesRegex(
            myokit.IntegrityError, 'the RHS still depends on <z.e.f>.',
            m.get('z.e').remove_child_variables)
        self.assertTrue(m.is_similar(m_org, True))

        # Remove children on nested var with used children
        self.assertRaisesRegex(
            myokit.IntegrityError,
            'depends on <z.e.f.h.[i|j]> and <z.e.f.h.[j|i]>.',
            m.get('z.e.f.h').remove_child_variables)
        self.assertTrue(m.is_similar(m_org, True))

        # Remove children on nested var with unused children
        e = m.get('z.e')
        h = e.get('f.h')
        self.assertEqual(len(list(h.variables())), 2)
        self.assertEqual(len(list(e.variables(deep=True))), 5)
        h.set_rhs('3 [m]')
        h.remove_child_variables()
        self.assertEqual(len(list(h.variables())), 0)
        self.assertEqual(len(list(e.variables(deep=True))), 3)
        m.validate()
        del h

        # Remove children on var with unused children
        e.set_rhs('1.234 [m]')
        self.assertEqual(len(list(e.variables(deep=True))), 3)
        e.remove_child_variables()
        self.assertEqual(len(list(e.variables(deep=True))), 0)
        m.validate()

    def test_rename(self):
        # Test :meth:`Variable.rename().

        # The functional part of this is done by Component.move_variable, so no
        # extensive testing is required
        m = myokit.Model()
        c = m.add_component('c')
        v = c.add_variable('v')
        v.rename('w')
        self.assertEqual(v.name(), 'w')
        self.assertEqual(v.qname(), 'c.w')

    def test_set_initial_value(self):
        # Test :meth:`Variable.set_initial_value()`.

        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        v = c.add_variable('v')
        v.set_rhs(3)
        w = c.add_variable('w')
        w.set_rhs('1 + 2')
        m.validate()

        # Test basic functionality
        v.promote(10)
        self.assertEqual(v.initial_value(), myokit.Number(10))
        v.set_initial_value(myokit.Number(12, myokit.units.g))
        self.assertEqual(v.initial_value(), myokit.Number(12, myokit.units.g))

        # Test setting expressions
        v.set_initial_value('1 + 11')
        #self.assertEqual(v.initial_value(),
        #                 myokit.Plus(myokit.Number(1), myokit.Number(11)))
        v.set_initial_value('1 + c.w')
        self.assertEqual(v.initial_value(),
                         myokit.Plus(myokit.Number(1), myokit.Name(w)))

        # Strings must use global context
        self.assertRaisesRegex(myokit.ParseError, 'No component specified for',
                               v.set_initial_value, '1 / w')

        # Only states have this option
        v.demote()
        self.assertRaisesRegex(
            Exception, 'Only state variables', v.set_initial_value, 3)
        self.assertRaisesRegex(
            Exception, 'Only state variables', w.set_initial_value, 3)
        v.promote(3)
        m.validate()

        # Initial values must be constant, but this is only picked up in
        # validation
        w.promote(1)
        v.set_initial_value(myokit.Name(w))
        self.assertRaises(myokit.IntegrityError, m.validate)
        w.demote()
        m.validate()

        # Nested variables may not be used, but this is only picked up when
        # validating
        x = v.add_variable('x')
        x.set_rhs(1)
        v.set_initial_value(x.lhs())
        self.assertRaises(
            myokit.IllegalReferenceInInitialValueError, m.validate)
        v.set_initial_value(x)

        # InitialValue and PartialDerivative can not be used
        v.set_initial_value(myokit.PartialDerivative(v.lhs(), x.lhs()))

        # Deprecated alias
        with WarningCollector() as w:
            v.set_state_value(1)
        self.assertIn('deprecated', w.text())
        self.assertEqual(v.initial_value(), myokit.Number(1))

    def test_set_unit(self):
        # Test :meth:`Variable.set_unit()`.
        m = myokit.Model()
        c = m.add_component('c')
        v = c.add_variable('v')

        # Test basic functionality
        s = myokit.UNIT_STRICT
        self.assertIsNone(v.unit())
        self.assertEqual(v.unit(s), myokit.units.dimensionless)
        v.set_unit(myokit.units.Newton)
        self.assertEqual(v.unit(), myokit.units.Newton)
        self.assertEqual(v.unit(s), myokit.units.Newton)

        # Set via unit parsing
        v.set_unit('kg/ms')
        self.assertEqual(v.unit(), myokit.parse_unit('kg/ms'))

        # Set to a non unit
        self.assertRaisesRegex(
            TypeError, 'expects a myokit.Unit', v.set_unit, 12)

    def test_unit(self):
        # Test :meth:`Variable.unit()`.
        m = myokit.Model()
        c = m.add_component('c')
        v = c.add_variable('v')
        d = myokit.units.dimensionless

        # Test no unit case
        self.assertIsNone(v.unit())
        self.assertEqual(v.unit(myokit.UNIT_STRICT), d)

        # RHS unit is not variable unit
        v.set_rhs('1 [ms]')
        self.assertIsNone(v.unit())
        self.assertEqual(v.unit(myokit.UNIT_STRICT), d)

        # Test unit-set case
        kg = myokit.units.kg
        v.set_unit(kg)
        self.assertEqual(v.unit(), kg)
        self.assertEqual(v.unit(myokit.UNIT_STRICT), kg)

    def test_validate(self):
        # Test some edge cases for validation.

        # Test scope rule:
        # Variables are allowed access all children of their ancestors
        m = myokit.Model()
        c = m.add_component('c')
        p = c.add_variable('p')
        p11 = p.add_variable('p11')
        p12 = p.add_variable('p12')
        p121 = p12.add_variable('p121')
        p121.set_rhs('p + p11 + p12')
        p121.validate()

        # But not children of children of ancestors
        p112 = p11.add_variable('p112')
        p121.set_rhs(p112.lhs())
        self.assertRaises(myokit.IllegalReferenceError, p121.validate)

        # RHS cannot be a partial, init, or condition
        m = myokit.Model()
        c = m.add_component('c')
        p = c.add_variable('p')
        p.set_rhs(3)
        q = c.add_variable('q')
        q.set_rhs(2)
        p.validate()
        p.set_rhs(myokit.PartialDerivative(p.lhs(), q.lhs()))
        self.assertRaisesRegex(
            myokit.IntegrityError, 'Partial derivatives', p.validate)
        p.promote(1)
        p.set_rhs(myokit.InitialValue(myokit.Name(p)))
        self.assertRaisesRegex(
            myokit.IntegrityError, 'Initial value', p.validate)
        p.set_rhs('q == 3')
        self.assertRaisesRegex(
            myokit.IntegrityError, 'can not be a condition', p.validate)
        p.set_rhs(3)
        p.validate()

        # Initial value can not reference nested variables
        r = c.add_variable('r')
        r1 = r.add_variable('r1')
        r1.set_rhs(1)
        r.set_rhs('1 + r1')
        p.validate()
        p.set_initial_value(myokit.Plus(myokit.Number(1), myokit.Name(r1)))
        self.assertRaises(
            myokit.IllegalReferenceInInitialValueError, p.validate)

        # Initial value must be constant
        p.set_initial_value('1 / c.q')
        q.promote(3)
        self.assertRaisesRegex(myokit.IntegrityError, 'not const', p.validate)
        q.demote()
        p.validate()
        q.set_rhs('1 + c.p')
        self.assertRaisesRegex(myokit.IntegrityError, 'not const', p.validate)
        q.set_rhs(10)
        p.validate()
        q.set_binding('hello')
        self.assertRaisesRegex(myokit.IntegrityError, 'not const', p.validate)
        # Don't test everything, is_const() should work!

        # Initial value cannot be a partial, init, or condition
        p.set_initial_value(myokit.PartialDerivative(p.lhs(), q.lhs()))
        self.assertRaisesRegex(
            myokit.IntegrityError, 'Partial derivatives', p.validate)
        p.set_initial_value(myokit.InitialValue(myokit.Name(p)))
        self.assertRaisesRegex(
            myokit.IntegrityError, 'Initial value', p.validate)
        p.set_initial_value('c.q == 3')
        self.assertRaisesRegex(
            myokit.IntegrityError, 'can not be a condition', p.validate)

        # Initial value cycles are caught in validation, don't cause loops
        m = myokit.parse_model('''
            [[model]]
            x.x = 0
            [x]
            dot(x) = x
            t = 0 bind time
            ''')
        x = m.get('x.x')
        x.set_initial_value('x.x')
        self.assertRaisesRegex(myokit.IntegrityError, 'not const', m.validate)
        x.set_initial_value(1)
        m.validate()
        x.set_initial_value('1 + 1 / x.x')
        self.assertRaisesRegex(myokit.IntegrityError, 'not const', m.validate)

    def test_value(self):
        # Test :meth:`Variable.value()`.

        m = myokit.Model()
        c = m.add_component('c')
        v = c.add_variable('v')
        v.set_rhs('1 + 2 + 3 + 4')
        self.assertEqual(v.value(), 10)


if __name__ == '__main__':
    unittest.main()
